"""
Mix-ins for building your own
`Django Datatable View <https://github.com/pivotal-energy-solutions/django-datatable-view>`_
powered views.
"""

__all__ = ['MultiTablesMixin']


class MultiTablesMixin(object):
    """Implements the base code for using multiple django-datatable-views powered tables."""

    multi_default = 'default'
    multi_datatables = ()
    request_name_key = 'datatable-name'

    def get_ajax_url(self, name=None):
        return self.request.path + f'?{self.request_name_key}={self.get_datatable_name(name)}'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        default_table = context.pop('datatable')
        context['datatables'] = [
            (
                name,
                label,
                (default_table if name == self.multi_default else self.get_datatable(name=name)))
            for name, label in self.multi_datatables
        ]
        return context

    def get_datatable(self, name=None):
        return self.get_datatable_structure(name=name)

    def get_datatable_name(self, name):
        return name or self.request.GET.get(self.request_name_key) or self.multi_default

    def get_datatable_structure(self, name=None):
        options = self._get_datatable_options()
        return self.datatable_structure_class(
            self.get_ajax_url(name=name),
            options,
            model=self.get_model())

    def get_queryset(self, name=None):
        qs = super().get_queryset()
        name = self.get_datatable_name(name)
        if name not in set(n for n, l in self.multi_datatables):
            return qs.none()  # Bad name, returns nothing!
        return getattr(self, f'get_{name}_queryset')(qs)



import os, tempfile

import cv2, numpy as np

from pytoolbox.network.http import download_ext


def load_image(path):
    """Reverse channels because OpenCV loads images in BGR mode."""
    return cv2.imread(path, 1)[..., ::-1]  # pylint:disable=no-member


def load_to_file(uri):
    if uri.startswith('http'):
        path = os.path.join(tempfile.gettempdir(), os.path.basename(uri))
        download_ext(uri, path, force=False)
        return path
    return uri


def normalize_rgb(image):
    """Scale integer RGB values [0,255] to float32 [0.0,1.0]."""
    return (image / 255).astype(np.float32)



import enum

from . import tag

__all__ = ['Image', 'Orientation']


class Orientation(enum.Enum):
    NORMAL = 1
    HOR_FLIP = 2
    ROT_180_CCW = 3
    VERT_FLIP = 4
    HOR_FLIP_ROT_270_CW = 5
    ROT_90_CW = 6
    HOR_FLIP_ROT_90_CW = 7
    ROT_270_CW = 8


class Image(tag.TagSet):

    ORIENTATION_TO_ROTATION = {
        None: 0,
        Orientation.NORMAL: 0,
        # 2 = Mirror horizontal
        Orientation.ROT_180_CCW: 180,
        # 4 = Mirror vertical
        # 5 = Mirror horizontal and rotate 270 CW
        Orientation.ROT_90_CW: -90,
        # 7 = Mirror horizontal and rotate 90 CW
        Orientation.ROT_270_CW: -270
    }

    def __init__(self, metadata, orientation=None):
        super().__init__(metadata)
        self._orientation = None if orientation is None else Orientation(orientation)

    @property
    def copyright(self):
        return self.metadata['Iptc.Application2.Copyright'].data

    @property
    def description(self):
        return self.metadata['Exif.Image.ImageDescription'].data

    @property
    def height(self):
        return self.clean_number(self.metadata.exiv2.get_pixel_height())

    @property
    def orientation(self):
        if self._orientation is not None:
            return self._orientation
        data = self.metadata['Exif.Image.Orientation'].data
        try:
            return Orientation(data)
        except Exception:  # pylint:disable=broad-except
            return None

    @property
    def rotation(self):
        return self.ORIENTATION_TO_ROTATION[self.orientation]

    @property
    def width(self):
        return self.clean_number(self.metadata.exiv2.get_pixel_width())



"""
Mix-ins for building your own `Django REST Framework <https://github.com/tomchristie/django-rest-framework>`_
powered API `serializers <http://www.django-rest-framework.org/tutorial/1-serialization/>`_.
"""

from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import smart_text
from django.utils.translation import gettext_lazy as _

from pytoolbox.django.models import utils

__all__ = ['BaseModelMixin', 'FromPrivateKeyMixin', 'NestedWriteMixin', 'ReadOnlyMixin']


class BaseModelMixin(object):

    def build_url_field(self, field_name, model_class):
        return super().build_url_field(field_name, utils.get_base_model(model_class))


class FromPrivateKeyMixin(object):
    """
    Allow to provide the PK of the model to retrieve it instead of creating a new instance with
    fields from data.
    """

    default_error_messages = {
        'does_not_exist': _('Invalid pk "{pk_value}" - object does not exist.'),
        'incorrect_type': _('Incorrect type. Expected pk value, received {data_type}.'),
    }

    def to_internal_value(self, data):
        """Transform the *incoming* primitive data into a native value."""
        if not data or isinstance(data, dict):
            return super().to_internal_value(data)
        try:
            return self.Meta.model.objects.get(pk=data)
        except ObjectDoesNotExist:
            self.fail('does_not_exist', pk_value=smart_text(data))
        except (TypeError, ValueError):
            self.fail('incorrect_type', data_type=type(data).__name__)

    def create(self, validated_data):
        if isinstance(validated_data, self.Meta.model):
            return validated_data
        return super().create(validated_data)


class NestedWriteMixin(object):

    def to_internal_value(self, data):
        """
        Return a tuple with (self, validate_data) to allow working on validated data with this
        serializer.
        """
        return self, super().to_internal_value(data)


class ReadOnlyMixin(object):

    def __init__(self, *args, **kwargs):
        kwargs['read_only'] = True
        super().__init__(*args, **kwargs)

    def create(self, validated_data):
        raise AttributeError('Read-only serializer')

    def update(self, task, validated_data):
        raise AttributeError('Read-only serializer')



import datetime, re, time

from pytoolbox.datetime import datetime_now, multiply_time, secs_to_time, str_to_time, time_ratio
from . import ffprobe, utils

__all__ = ['ENCODING_REGEX', 'EncodeState', 'EncodeStatistics', 'FrameBasedRatioMixin']

ENCODING_REGEX = re.compile(
    # frame= 2071 fps=  0 q=-1.0 size=   34623kB time=00:01:25.89 bitrate=3302.3kbits/s
    r'frame=\s*(?P<frame>\d+)\s+'
    r'fps=\s*(?P<frame_rate>\d+\.?\d*)\s+'
    r'q=\s*(?P<qscale>\S+)\s+\S*.*'
    r'size=\s*(?P<size>\S+)\s+'
    r'time=\s*(?P<time>\S+)\s+'
    r'bitrate=\s*(?P<bit_rate>\S+)'
)


class EncodeState(object):
    NEW = 'NEW'
    STARTED = 'STARTED'
    PROCESSING = 'PROCESSING'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'

    ALL_STATES = frozenset([NEW, STARTED, PROCESSING, SUCCESS, FAILURE])
    FINAL_STATES = frozenset([SUCCESS, FAILURE])


class EncodeStatistics(object):  # pylint:disable=too-many-instance-attributes

    default_in_duration = datetime.timedelta(seconds=0)
    encoding_regex = ENCODING_REGEX
    ffprobe_class = ffprobe.FFprobe
    states = EncodeState

    def __init__(self, inputs, outputs, in_options, out_options, in_base_index=0, out_base_index=0):
        self.inputs = inputs
        self.outputs = outputs
        self.in_options = in_options
        self.out_options = out_options
        self.in_base_index = in_base_index
        self.out_base_index = out_base_index

        self.process = None
        self.process_output = ''
        self.returncode = None

        self.state = self.states.NEW
        self.start_date = None
        self.start_time = None
        self.elapsed_time = None
        self.ratio = None
        self.frame = None
        self.frame_rate = None
        self.qscale = None
        self.bit_rate = None

        # Retrieve input media duration and size, handle sub-clipping
        duration = self.ffprobe_class().get_media_duration(self.input, as_delta=True)
        duration = duration or self.default_in_duration
        self.input.duration, self.input.size = \
            self._get_subclip_duration_and_size(duration, self.input.size, self.out_options)
        self.output.duration = None

    @property
    def eta_time(self):
        if self.state in self.states.FINAL_STATES:
            return datetime.timedelta(0)
        if not self.ratio:
            return None
        return multiply_time(self.elapsed_time, (1.0 - self.ratio) / self.ratio, as_delta=True)

    @property
    def input(self):
        return self.inputs[self.in_base_index]

    @property
    def output(self):
        return self.outputs[self.out_base_index]

    @staticmethod
    def get_now():
        return datetime_now()

    def start(self, process):
        self.state = self.states.STARTED
        self.process = process
        self.start_date = self.get_now()
        self.start_time = time.time()
        self.elapsed_time = datetime.timedelta(0)
        self.output.duration = datetime.timedelta(0)
        self.frame = 0
        self._update_ratio()
        return self

    def progress(self, chunk):
        self.state = self.states.PROCESSING
        self.elapsed_time = datetime.timedelta(seconds=time.time() - self.start_time)
        ffmpeg_statistics = self._parse_chunk(chunk)
        if ffmpeg_statistics:
            self.output.duration = ffmpeg_statistics['time']
            self.frame = ffmpeg_statistics['frame']
            self.frame_rate = ffmpeg_statistics['frame_rate']
            self.qscale = ffmpeg_statistics['qscale']
            self.output.size = ffmpeg_statistics['size']
            self.bit_rate = ffmpeg_statistics['bit_rate']
        self._update_ratio()
        return self

    def end(self, returncode):
        self.state = self.states.FAILURE if returncode else self.states.SUCCESS
        self.returncode = returncode
        self.elapsed_time = datetime.timedelta(seconds=time.time() - self.start_time)
        self.frame_rate = self.frame / (self.elapsed_time.total_seconds() or 0.0001)
        self.output.duration = self.ffprobe_class().get_media_duration(
            self.output.path, as_delta=True)
        self.output.size = None
        self._update_ratio()
        return self

    def _update_ratio(self):
        if self.state == self.states.SUCCESS:
            self.ratio = 1.0
        else:
            ratio = self._compute_ratio()
            self.ratio = None if ratio is None else min(1.0, max(0.0, ratio))

    def _compute_ratio(self):
        if self.input.duration and self.output.duration is not None:
            return time_ratio(self.output.duration, self.input.duration)
        return None

    @classmethod
    def _get_subclip_duration_and_size(cls, duration, size, out_options):
        """Adjust duration and size if we only encode a sub-clip."""

        try:
            sub_duration = cls._to_time(out_options[out_options.index('-t') + 1])
        except (IndexError, ValueError):
            sub_duration = duration
        if sub_duration is None:
            return duration, size

        zero = datetime.timedelta(0)
        try:
            sub_position = cls._to_time(out_options[out_options.index('-ss') + 1]) or zero
        except (IndexError, ValueError):
            sub_position = zero

        sub_duration = max(zero, min(duration - sub_position, sub_duration))
        return sub_duration, int(size * time_ratio(sub_duration, duration))

    def _parse_chunk(self, chunk):
        self.process_output += chunk
        match = self.encoding_regex.match(chunk.strip())
        if not match:
            return None
        ffmpeg_statistics = match.groupdict()
        try:
            ffmpeg_statistics['time'] = str_to_time(ffmpeg_statistics['time'], as_delta=True)
        except ValueError:
            return None  # Parsed statistics are broken, do not use them
        ffmpeg_statistics['frame'] = int(ffmpeg_statistics['frame'])
        ffmpeg_statistics['frame_rate'] = float(ffmpeg_statistics['frame_rate'])
        qscale = ffmpeg_statistics.get('qscale')
        ffmpeg_statistics['qscale'] = None if qscale is None else float(qscale)
        ffmpeg_statistics['size'] = utils.to_size(ffmpeg_statistics['size'])
        ffmpeg_statistics['bit_rate'] = utils.to_bit_rate(ffmpeg_statistics['bit_rate'])
        return ffmpeg_statistics

    @staticmethod
    def _to_time(value):
        method = str_to_time if ':' in value else secs_to_time
        return method(value, as_delta=True)


class FrameBasedRatioMixin(object):
    """
    Compute ratio based on estimated input number of frames and current output number of frames.
    Fall-back to super's ratio computation method.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fps = self.ffprobe_class().get_video_frame_rate(self.input)
        if fps and self.input.duration:
            self.input.frame = fps * self.input.duration.total_seconds()
        else:
            self.input.frame = None

    def _compute_ratio(self):
        if self.input.frame and self.frame is not None:
            return self.frame / self.input.frame
        return super()._compute_ratio()



import os, re

from . import module

_all = module.All(globals())

ALL_CAP_REGEX = re.compile(r'([a-z0-9])([A-Z])')
FIRST_CAP_REGEX = re.compile(r'(.)([A-Z][a-z]+)')


def camel_to_dash(string):
    """Convert camelCase to dashed-case."""
    sub_string = FIRST_CAP_REGEX.sub(r'\1-\2', string)
    dashed_case_str = ALL_CAP_REGEX.sub(r'\1-\2', sub_string).lower()
    return dashed_case_str.replace('--', '-')


def camel_to_snake(string):
    """Convert camelCase to snake_case."""
    sub_string = FIRST_CAP_REGEX.sub(r'\1_\2', string)
    snake_cased_str = ALL_CAP_REGEX.sub(r'\1_\2', sub_string).lower()
    return snake_cased_str.replace('__', '_')


def dash_to_camel(string):
    return _to_camel(string, '-')


def snake_to_camel(string):
    return _to_camel(string, '_')


def _to_camel(string, separator):
    components = string.split(separator)
    preffix = suffix = ''
    if components[0] == '':
        components = components[1:]
        preffix = separator
    if components[-1] == '':
        components = components[:-1]
        suffix = separator
    if len(components) > 1:
        camel_case_string = components[0].lower()
        for component in components[1:]:
            if component.isupper() or component.istitle():
                camel_case_string += component
            else:
                camel_case_string += component.title()
    else:
        camel_case_string = components[0]
    return preffix + camel_case_string + suffix


def filterjoin(items, sep=' ', keep=lambda o: o):
    """
    Concatenate `items` with intervening occurrences of `sep`. Gracefully convert items to string
    and filter the items using the `keep` function.
    """
    return sep.join(str(i) for i in items if keep(i))


def to_lines(items, limit=80, start='\t', template='{0} '):
    """
    Convert items to string using `template`.
    Ensure lines length of maximum `limit`.
    Prefixing every line with `start`.

    **Example usage***

    >>> some_culture = (  # Credits: https://en.wikipedia.org/wiki/Punched_card
    ...     "A legacy of the 80 column punched card format is that a display of 80 characters per"
    ...     " row was a common choice in the design of character-based terminals. As of November"
    ...     " 2011 some character interface defaults, such as the command prompt window's width"
    ...     " in Microsoft Windows, remain set at 80 columns and some file formats, such as FITS,"
    ...     " still use 80-character card images.")

    Using default options:

    >>> print(to_lines(some_culture.split(' ')))
        A legacy of the 80 column punched card format is that a display of 80
        characters per row was a common choice in the design of character-based
        terminals. As of November 2011 some character interface defaults, such as the
        command prompt window's width in Microsoft Windows, remain set at 80 columns
        and some file formats, such as FITS, still use 80-character card images.

    Customizing output:

    >>> print(to_lines(some_culture.split(' '), limit=42, start='> '))
    > A legacy of the 80 column punched card
    > format is that a display of 80
    > characters per row was a common choice
    > in the design of character-based
    > terminals. As of November 2011 some
    > character interface defaults, such as
    > the command prompt window's width in
    > Microsoft Windows, remain set at 80
    > columns and some file formats, such as
    > FITS, still use 80-character card
    > images.

    Displaying objects representation:

    >>> class Item(object):
    ...     def __init__(self, value):
    ...         self.value = value
    ...
    ...     def __repr__(self):
    ...         return f'<Item value={self.value}>'
    ...
    >>> print(to_lines((Item(i) for i in range(22)), limit=60, template='{0!r} '))
        <Item value=0> <Item value=1> <Item value=2>
        <Item value=3> <Item value=4> <Item value=5>
        <Item value=6> <Item value=7> <Item value=8>
        <Item value=9> <Item value=10> <Item value=11>
        <Item value=12> <Item value=13> <Item value=14>
        <Item value=15> <Item value=16> <Item value=17>
        <Item value=18> <Item value=19> <Item value=20>
        <Item value=21>
    """
    lines = [start]
    for item in items:
        item_str = template.format(item)
        if len(lines[-1]) + len(item_str) > limit:
            lines.append(start)
        lines[-1] += item_str
    return os.linesep.join(lines)


__all__ = _all.diff(globals())



import fnmatch, re

from .itertools import chain

__all__ = [
    'TIME_REGEX_PARTS',
    'UUID_REGEX',
    'embed_in_regex',
    'findall_partial',
    'from_path_patterns'
]

TIME_REGEX_PARTS = ['[0-2]', '[0-9]', ':', '[0-5]', '[0-9]', ':', '[0-5]', '[0-9]']
UUID_REGEX = '[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}'


def embed_in_regex(string, regex_parts, index, as_string=True):
    """
    **Example usage**

    >>> embed_in_regex('L', ['[a-z]', '[a-z]'], 0)
    (0, 'L[a-z]')
    >>> embed_in_regex('L', ['[a-z]', '[a-z]'], 1)
    (1, '[a-z]L')
    >>> embed_in_regex('L', ['[a-z]', '[a-z]'], 1, as_string=False)
    (1, ['[a-z]', 'L'])
    """
    regex = regex_parts[:]
    regex[index:index + len(string)] = string
    return index, ''.join(regex) if as_string else regex


def findall_partial(string, regex_parts):
    """
    **Example usage**

    >>> [i for s, r, i in findall_partial(':', TIME_REGEX_PARTS)]
    [2, 5]
    >>> [embed_in_regex(s, r, i) for s, r, i in findall_partial('12:15:2', TIME_REGEX_PARTS)]
    [(0, '12:15:2[0-9]')]
    >>> [embed_in_regex(s, r, i) for s, r, i in findall_partial('18:2', TIME_REGEX_PARTS)]
    [(0, '18:2[0-9]:[0-5][0-9]'), (3, '[0-2][0-9]:18:2[0-9]')]
    >>> [embed_in_regex(s, r, i) for s, r, i in findall_partial('59:1', TIME_REGEX_PARTS)]
    [(3, '[0-2][0-9]:59:1[0-9]')]
    """
    for index in range(0, len(regex_parts) - len(string) + 1):  # noqa
        regex = regex_parts[index:index + len(string)]
        match = re.search(''.join(regex), string)
        if match:
            yield string, regex_parts, index


def from_path_patterns(patterns, regex=False):
    """
    Return patterns compiled to regular expressions, if necessary.

    If `regex` is set to False, then any string pattern will be converted from the unix-style
    wildcard to the regular expression equivalent using :func:`fnatmch.translate`.

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>> asserts.list_equal(
    ...     [r.pattern for r in from_path_patterns('*.txt')],
    ...     ['(?s:.*\\\\.txt)\\\\Z'])
    >>> asserts.list_equal(
    ...     [r.pattern for r in from_path_patterns(['text', 'something*'], '*.py')],
    ...     ['text', 'something*'])
    >>> asserts.list_equal(
    ...     [r.pattern for r in from_path_patterns([re.compile('a?c'), 'fo?'])],
    ...     ['a?c', '(?s:fo.)\\\\Z'])
    >>> asserts.list_equal(
    ...     [r.pattern for r in from_path_patterns([re.compile('a?c'), 'foo?'], regex=True)],
    ...     ['a?c', 'foo?'])
    """
    return [
        p if hasattr(p, 'match') else re.compile(p if regex else fnmatch.translate(p))
        for p in chain(patterns)
    ]


class Match(object):
    """
    Assert that a given string meets some expectations.

    Credits: https://kalnytskyi.com/howto/assert-str-matches-regex-in-pytest/
    """

    def __init__(self, pattern, flags=0):
        self._regex = re.compile(pattern, flags)

    def __eq__(self, actual):
        return bool(self._regex.match(actual))

    def __repr__(self):
        return self._regex.pattern



import datetime

from pytoolbox import exceptions
from . import camera, image, lens, photo, tag

__all__ = ['Metadata']


class Metadata(object):

    camera_class = camera.Camera
    image_class = image.Image
    lens_class = lens.Lens
    photo_class = photo.Photo
    tag_class = tag.Tag

    def __init__(self, path=None, buf=None, orientation=None, gexiv2_version='0.10'):
        import gi  # pylint:disable=import-error
        gi.require_version('GExiv2', gexiv2_version)
        from gi.repository import GExiv2  # pylint:disable=import-error
        self.path = path
        self.exiv2 = GExiv2.Metadata()
        if buf:
            self.exiv2.open_buf(buf)
        elif path:
            self.exiv2.open_path(path)
        else:
            raise ValueError('buf or file is required')
        self.camera = self.camera_class(self)
        self.image = self.image_class(self, orientation)
        self.lens = self.lens_class(self)
        self.photo = self.photo_class(self)

    def __getitem__(self, key):
        # TODO make it more strict and re-implement less strict self.get(key)
        return self.tag_class(self, key)

    def __setitem__(self, key, value):
        self.exiv2[key] = value

    @property
    def tags(self):
        return {k: self[k] for k in self.exiv2.get_tags()}

    def get_date(self, keys=('Exif.Photo.DateTimeOriginal', 'Exif.Image.DateTime')):
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            date = self[key].data
            if isinstance(date, datetime.datetime):
                return date
        return None

    def rewrite(self, path=None, save=False):
        """
        Iterate over all tags and rewrite them to fix issues (e.g. GExiv2: Invalid ifdId 103 (23)).
        """
        tags = {k: str(v.data) for k, v in self.tags.items()}
        self.exiv2.clear()
        for key, value in tags.items():
            self[key] = value
        if save:
            self.save_file(path=path)

    def save_file(self, path=None):
        if not path and not self.path:
            raise exceptions.UndefinedPathError()
        return self.exiv2.save_file(path=path or self.path)



import collections, math

from . import module
from .datetime import total_seconds

_all = module.All(globals())


class EventsTable(object):
    """Scan a spare events table and replace missing entry by previous (non empty) entry."""

    def __init__(self, sparse_events_table, time_range, time_speedup, sleep_factor=1.0):
        self.time_range = time_range
        self.time_speedup = time_speedup
        self.sleep_factor = sleep_factor
        previous_event = sparse_events_table[0]
        self.events = {}
        for index in range(self.time_range):  # noqa
            event = sparse_events_table.get(index, previous_event)
            self.events[index] = event
            previous_event = event

    def get(self, time, time_speedup=None, default_value=None):
        # """
        # >>> def test_get_index(time_range, time_speedup):
        # ...     table = EventsTable({0: 'salut'}, time_range, time_speedup)
        # ...     modulo = previous = 0
        # ...     for t in range(24 * 3600 + 1):
        # ...         index = table.get(t+2*60, time_range, time_speedup)[0]
        # ...         if previous > index:
        # ...             modulo += 1
        # ...         assert 0 <= index < time_range
        # ...         previous = index
        # ...     assert modulo == time_speedup

        # Test get_index with a speedup of 1440 (maps 1 minute to 24 hours):
        # >>> test_get_index(24, 24 * 60)

        # Test get_index with a speedup of 12 (maps 2 hours to 24 hours):
        # >>> test_get_index(24, 12)
        # """
        speedup = time_speedup or self.time_speedup
        index = int(total_seconds(time) * (speedup / 3600) % self.time_range)
        return index, self.events.get(index, default_value)

    def sleep_time(self, time, time_speedup=None, sleep_factor=None):
        """
        Return required sleep time to wait for next scheduled event.

        **Example usage**

        >>> table = EventsTable({0: 'salut'}, 24, 60)
        >>> table.sleep_time(1)
        59
        >>> table.sleep_time(58)
        2
        >>> table.sleep_time(60)
        60
        >>> table.sleep_time(62)
        58
        >>> table.sleep_time(3590, time_speedup=1)
        10
        >>> table.sleep_time(12543, time_speedup=1)
        1857
        >>> table.sleep_time(12543, time_speedup=1, sleep_factor=2)
        57
        >>> table.sleep_time(12600, time_speedup=1, sleep_factor=2)
        1800
        >>> table.sleep_time(1, time_speedup=60, sleep_factor=1)
        59
        >>> table.sleep_time(1, time_speedup=60, sleep_factor=2)
        29
        >>> table.sleep_time(30, time_speedup=60, sleep_factor=2)
        30
        >>> table.time_range = 1
        >>> table.sleep_time(1, time_speedup=1)
        149
        """
        # 150 = 3600 / 24
        speedup = time_speedup or self.time_speedup
        factor = sleep_factor or self.sleep_factor
        duration = self.time_range * 150 / (speedup * factor)
        return math.ceil(duration - (total_seconds(time) % duration))


class pygal_deque(collections.deque):  # pylint:disable=invalid-name
    """
    A deque None'ing duplicated values to produce nicer `pygal <pygal.org>`_ line charts
    (e.g. 5555322211111 -> 5__532_21___1).

    .. warning::

        Not yet implemented:

        * appendleft(x)
        * clear()
        * count(x)
        * extend(iterable)
        * extendleft(iterable)
        * pop()
        * popleft()
        * remove(value)
        * reverse()
        * rotate(n)
    """

    last = None

    def append(self, value):  # pylint:disable=arguments-renamed
        if value != self.last and value is not None:
            try:
                self[-1] = self.last
            except Exception:  # pylint:disable=broad-except
                pass
            self.last = value
        else:
            value = None
        try:
            first = self[0]
        except IndexError:
            first = None
        super().append(value)
        if self[0] is None:
            self[0] = first

    def list(self, fill=False):
        self_list = list(self)
        try:
            if self.last is not None:
                self_list[-1] = self.last
        except Exception:  # pylint:disable=broad-except
            pass
        if fill and self_list:
            previous = None
            for index, _ in enumerate(self_list):
                if self_list[index] is None:  # pylint:disable=unnecessary-list-index-lookup
                    self_list[index] = previous
                else:
                    previous = self_list[index]
        return self_list


def flatten_dict(the_dict, key_template='{0}.{1}'):
    """
    Flatten the keys of a nested dictionary. Nested keys will be appended iteratively using given
    `key_template`.

    **Example usage**

    >>> flatten_dict({'a': 'b', 'c': 'd'})
    {'a': 'b', 'c': 'd'}
    >>> flatten_dict({'a': {'b': {'c': ['d', 'e']}, 'f': 'g'}})
    {'a.b.c': ['d', 'e'], 'a.f': 'g'}
    >>> flatten_dict({'a': {'b': {'c': ['d', 'e']}, 'f': 'g'}}, '{1}-{0}')
    {'c-b-a': ['d', 'e'], 'f-a': 'g'}
    """
    def expand_item(key, value):
        if isinstance(value, dict):
            return [
                (key_template.format(key, k), v)
                for k, v in flatten_dict(value, key_template).items()
            ]
        return [(key, value)]
    return dict(item for k, v in the_dict.items() for item in expand_item(k, v))


def merge_dicts(*dicts):
    """
    Return a dictionary from multiple dictionaries.

    .. warning::

        This operation is not commutative.

    **Example usage**

    >>> merge_dicts({'a': 1, 'b': 2}, {'b': 3, 'c': 4}, {'c': 5})
    {'a': 1, 'b': 3, 'c': 5}
    >>> merge_dicts({'c': 5}, {'b': 3, 'c': 4}, {'a': 1, 'b': 2})
    {'c': 4, 'b': 2, 'a': 1}
    """
    merged_dict = {}
    for the_dict in dicts:
        merged_dict.update(the_dict)
    return merged_dict


def swap_dict_of_values(the_dict, type=set, method=set.add):  # pylint:disable=redefined-builtin
    """
    Return a dictionary (:class:`collections.defaultdict`) with keys and values swapped.

    This algorithm expect that the values are a container with objects, not a single object.
    Set type to None if values are unique and you want keys to be the values.

    **Example usage**

    Simple swap:

    >>> swap_dict_of_values({'odd': [1, 3], 'even': (0, 2)}, type=None)
    {1: 'odd', 3: 'odd', 0: 'even', 2: 'even'}

    Complex swap:

    >>> def S(value):
    ...     return {k: sorted(v) for k, v in sorted(value.items())}
    ...
    >>> S(swap_dict_of_values(
    ...     {'odd': [1, 3], 'even': (0, 2), 'fib': {1, 2, 3}},
    ...     type=list,
    ...     method=list.append))
    {0: ['even'], 1: ['fib', 'odd'], 2: ['even', 'fib'], 3: ['fib', 'odd']}
    >>> swap_dict_of_values({'o': [1, 3], 'e': (0, 2), 'f': {2, 3}}, method='add')[2] == {'f', 'e'}
    True
    >>> swap_dict_of_values({'bad': 'ab', 'example': 'ab'})['a'] == {'bad', 'example'}
    True
    """
    if type is None:
        reversed_dict = {}
        for key, values in the_dict.items():
            for value in values:
                reversed_dict[value] = key
    else:
        method = getattr(type, method) if isinstance(method, str) else method
        reversed_dict = collections.defaultdict(type)
        for key, values in the_dict.items():
            for value in values:
                method(reversed_dict[value], key)
    return reversed_dict


def to_dict_of_values(iterable, type=list, method=list.append):  # pylint:disable=redefined-builtin
    """
    Return a dictionary (:class:`collections.defaultdict`) with key, value pairs merged as
    key -> values.

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>> asserts.dict_equal(
    ...     to_dict_of_values([('odd', 1), ('odd', 3), ('even', 0), ('even', 2)]),
    ...     {'even': [0, 2], 'odd': [1, 3]})
    >>> asserts.dict_equal(
    ...     to_dict_of_values((('a', 1), ('a', 1), ('a', 2)), type=set, method=set.add),
    ...     {'a': {1, 2}})
    """
    dict_of_values = collections.defaultdict(type)
    for key, value in iterable:
        method(dict_of_values[key], value)
    return dict_of_values


def window(values, index, delta):
    """
    Extract 1+2*`delta` items from `values` centered at `index` and return a tuple with
    (items, left, right).

    This function tries to simulate a physical slider, meaning the number of extracted elements is
    constant but the centering at `index` is not guaranteed.

    A visual example with 6 `values` and ``delta=1``::

        index = 0  [+++]---  left = 0, right = 2
        index = 1  [+++]---  left = 0, right = 2
        index = 2  -[+++]--  left = 1, right = 3
        index = 3  --[+++]-  left = 2, right = 4
        index = 4  ---[+++]  left = 3, right = 5
        index = 5  ---[+++]  left = 3, right = 5

    **Example usage**

    >>> window(['a'], 0, 2)
    (['a'], 0, 0)
    >>> window(['a', 'b', 'c', 'd'], 2, 0)
    (['c'], 2, 2)
    >>> window(['a', 'b', 'c', 'd', 'e'], 0, 1)
    (['a', 'b', 'c'], 0, 2)
    >>> window(['a', 'b', 'c', 'd', 'e'], 1, 1)
    (['a', 'b', 'c'], 0, 2)
    >>> window(['a', 'b', 'c', 'd', 'e'], 2, 1)
    (['b', 'c', 'd'], 1, 3)
    >>> window(['a', 'b', 'c', 'd', 'e'], 3, 1)
    (['c', 'd', 'e'], 2, 4)
    >>> window(['a', 'b', 'c', 'd', 'e'], 4, 1)
    (['c', 'd', 'e'], 2, 4)
    >>> window(['a', 'b', 'c', 'd', 'e'], 3, 6)
    (['a', 'b', 'c', 'd', 'e'], 0, 4)
    >>> data = list(range(20))
    >>> window(data, 6, 6)
    ([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], 0, 12)
    >>> window(data, 7, 6)
    ([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13], 1, 13)
    >>> window(data, 10, 6)
    ([4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16], 4, 16)
    >>> window(data, 19, 6)
    ([7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19], 7, 19)
    """
    length = len(values)
    left, right = index - delta, index + delta
    if left < 0:
        left, right = 0, right - left
    elif right >= length:
        left, right = left - right + length - 1, length - 1
    left, right = max(left, 0), min(right, length - 1)
    return values[left:right + 1], left, right


__all__ = _all.diff(globals())



import struct

__all__ = ['RtpPacket']


class RtpPacket(object):  # pylint:disable=too-many-instance-attributes
    """
    This represent a real-time transport protocol (RTP) packet.

    * :rfc:`3550`
    * `Wikipedia (RTP) <http://en.wikipedia.org/wiki/Real-time_Transport_Protocol>`_
    * `Parameters (RTP) <http://www.iana.org/assignments/rtp-parameters/rtp-parameters.xml>`_

    **Packet header**

    * RFC 3550 page 13

    .. code-block:: text

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |V=2|P|X|  CC   |M|     PT      |       sequence number         |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                           timestamp                           |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |           synchronization source (SSRC) identifier            |
       +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
       |            contributing source (CSRC) identifiers             |
       |                             ....                              |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    **Extension header**

    * RFC 3550 page 19

    .. code-block:: text

         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |      defined by profile       |           length              |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                        header extension                       |
        |                             ....                              |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constants >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    ER_VERSION = 'RTP Header : Version must be set to 2'
    ER_PADDING_LENGTH = 'RTP Header : Bad padding length'
    ER_EXTENSION_LENGTH = 'RTP Header : Bad extension length'
    ER_PAYLOAD = 'RTP packet must have a payload'

    HEADER_LENGTH = 12
    V_MASK = 0xc0
    V_SHIFT = 6
    P_MASK = 0x20
    X_MASK = 0x10
    CC_MASK = 0x0f
    M_MASK = 0x80
    PT_MASK = 0x7f
    DYNAMIC_PT = 96   # Dynamic payload type
    MP2T_PT = 33      # MPEG2 TS payload type
    MP2T_CLK = 90000  # MPEG2 TS clock rate [Hz]
    S_MASK = 0x0000ffff
    TS_MASK = 0xffffffff

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    @property
    def valid(self):
        """Returns True if this packet is a valid RTP packet."""
        return len(self.errors) == 0

    @property
    def validMP2T(self):  # pylint:disable=invalid-name
        """Returns True if this packet is a valid RTP packet containing a MPEG2-TS payload."""
        return self.valid and self.payload_type == self.MP2T_PT

    @property
    def errors(self):
        """
        Returns an array containing any errors.

        :return: array of error message(s).

        **Example usage**

        Testing invalid header:

        >>> from pytoolbox.unittest import asserts
        >>> rtp = RtpPacket(bytearray(RtpPacket.HEADER_LENGTH-1), RtpPacket.HEADER_LENGTH-1)
        >>> asserts.list_equal(rtp.errors, [
        ...     'RTP Header : Version must be set to 2',
        ...     'RTP packet must have a payload'
        ... ])

        Testing a valid RTP packet with a MPEG2-TS payload:

        >>> rtp = RtpPacket.create(6, 777, RtpPacket.MP2T_PT, 'salut')
        >>> asserts.list_equal(rtp.errors, [])
        """
        errors = []
        if self._error:
            errors.append(self._error)
        if self.version != 2:
            errors.append(self.ER_VERSION)
        if self.payload_size == 0:
            errors.append(self.ER_PAYLOAD)
        return errors

    @property
    def clock_rate(self):
        """Return the MPEG2-TS clock rate of a MPEG2-TS payload or 1 if this is not."""
        return self.MP2T_CLK if self.payload_type == self.MP2T_PT else 1

    @property
    def header_size(self):
        """
        Returns the length (aka size) of the header.

        **Example usage**

        >>> rtp = RtpPacket.create(6, 777, RtpPacket.MP2T_PT, 'salut')
        >>> print(rtp.header_size)
        12
        """
        return self.HEADER_LENGTH + 4 * len(self.csrc)

    @property
    def payload_size(self):
        """
        Returns the length (aka size) of the payload.

        **Example usage**

        >>> rtp = RtpPacket.create(6, 777, RtpPacket.MP2T_PT, 'salut')
        >>> print(rtp.payload_size)
        5
        """
        return len(self.payload) if self.payload else 0

    @property
    def time(self):
        """Return computed time (*timestamp / clock rate*)."""
        return self.timestamp / self.clock_rate

    @property
    def header_bytes(self):
        """
        Return the RTP header bytes.

        *Example usage*

        >>> rtp = RtpPacket.create(6, 777, RtpPacket.MP2T_PT, bytearray.fromhex('00 01 02 03'))
        >>> print(rtp)
        version      = 2
        errors       = []
        padding      = False
        extension    = False
        marker       = False
        payload type = 33
        sequence     = 6
        timestamp    = 777
        clock rate   = 90000
        time         = 0
        ssrc         = 0
        csrc count   = 0
        payload size = 4
        >>> header = rtp.header_bytes
        >>> assert len(header) == 12
        >>> print(''.join(' %02x' % b for b in header))
         80 21 00 06 00 00 03 09 00 00 00 00
        >>> header += rtp.payload
        >>> assert rtp == RtpPacket(header, len(header))

        >>> rtp = RtpPacket.create(0xffffffff, 0xffffffffff, RtpPacket.DYNAMIC_PT, bytearray(1023))
        >>> print(rtp)
        version      = 2
        errors       = []
        padding      = False
        extension    = False
        marker       = False
        payload type = 96
        sequence     = 65535
        timestamp    = 4294967295
        clock rate   = 1
        time         = 4294967295
        ssrc         = 0
        csrc count   = 0
        payload size = 1023
        >>> header = rtp.header_bytes
        >>> assert len(header) == 12
        >>> print(''.join(' %02x' % b for b in header))
         80 60 ff ff ff ff ff ff 00 00 00 00
        >>> header += rtp.payload
        >>> assert rtp == RtpPacket(header, len(header))
        """
        cc = len(self.csrc)  # pylint:disable=invalid-name
        header = bytearray(self.HEADER_LENGTH + 4 * cc)
        header[0] = (
            ((self.version << self.V_SHIFT) & self.V_MASK)
            + (self.P_MASK if self.padding else 0)
            + (self.X_MASK if self.extension else 0)
            + (cc & self.CC_MASK))
        header[1] = (
            (self.M_MASK if self.marker else 0)
            + (self.payload_type & self.PT_MASK))
        struct.pack_into(b'!H', header, 2, self.sequence)
        struct.pack_into(b'!I', header, 4, self.timestamp)
        struct.pack_into(b'!I', header, 8, self.ssrc)
        i = 12
        for contributor in self.csrc:
            struct.pack_into(b'!I', header, i, contributor)
            i += 4
        return header

    @property
    def bytes(self):
        """Return the RTP packet header and payload bytes."""
        return self.header_bytes + self.payload

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructor >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    def __init__(self, data, length):
        """
        This constructor will parse input bytes array to fill packet's fields.
        In case of error (e.g. bad version number) the constructor will abort filling
        fields and un-updated fields are set to their corresponding default value.

        :param bytes: Input array of bytes to parse as a RTP packet
        :type bytes: bytearray
        :param length: Amount of bytes to read from the array of bytes
        :type length: int

        **Example usage**

        Testing invalid headers:

        >>> rtp = RtpPacket(bytearray(RtpPacket.HEADER_LENGTH-1), RtpPacket.HEADER_LENGTH-1)
        >>> rtp.valid  # Bad length
        False
        >>> rtp = RtpPacket(bytearray(RtpPacket.HEADER_LENGTH), RtpPacket.HEADER_LENGTH)
        >>> rtp.valid  # Bad version
        False
        >>> bytes = bytearray(RtpPacket.HEADER_LENGTH)
        >>> bytes[0] = 0xa0
        >>> rtp = RtpPacket(bytes, RtpPacket.HEADER_LENGTH)
        >>> rtp.valid  # Padding enabled but not present
        False

        Testing header fields value:

        >>> bytes = bytes.fromhex('80 a1 a4 25 ca fe b5 04 b0 60 5e bb 12 34')
        >>> rtp = RtpPacket(bytes, len(bytes))
        >>> rtp.valid
        True
        >>> print(rtp)
        version      = 2
        errors       = []
        padding      = False
        extension    = False
        marker       = True
        payload type = 33
        sequence     = 42021
        timestamp    = 3405690116
        clock rate   = 90000
        time         = 37841
        ssrc         = 2959105723
        csrc count   = 0
        payload size = 2
        >>> rtp.csrc
        []
        >>> rtp.payload[0]
        18
        >>> rtp.payload[1]
        52

        Testing header fields value (with padding, extension and ccrc):

        >>> bytes = bytes.fromhex('b5a1a401 cafea421 b0605ebb 11111111 22222222 33333333 '
        ...                       '44444444 55555555 00000004 87654321 12340002')
        >>> rtp = RtpPacket(bytes, len(bytes))
        >>> rtp.valid
        True
        >>> rtp.version
        2
        >>> rtp.padding
        True
        >>> rtp.extension
        True
        >>> rtp.marker
        True
        >>> rtp.payload_type
        33
        >>> rtp.sequence
        41985
        >>> rtp.timestamp
        3405685793
        >>> rtp.clock_rate
        90000
        >>> rtp.ssrc
        2959105723
        >>> len(rtp.csrc)
        5
        >>> rtp.csrc
        [286331153, 572662306, 858993459, 1145324612, 1431655765]
        >>> rtp.payload
        bytearray(b'\\x124')
        """

        # Fields default values
        self.version = 0
        self.padding = False
        self.extension = False
        self.marker = False
        self.payload_type = 0
        self.sequence = 0
        self.timestamp = 0
        self.ssrc = 0
        self.csrc = []
        self.payload = []
        self._error = None

        offset = self.HEADER_LENGTH
        if length < offset:
            return

        self.version = (data[0] & self.V_MASK) >> 6
        if self.version != 2:
            return

        self.padding = (data[0] & self.P_MASK) == self.P_MASK
        if self.padding:  # Remove padding if present
            padding_length = data[-1]
            if padding_length == 0 or length < (offset + padding_length):
                self._error = self.ER_PADDING_LENGTH
                return
            length -= padding_length

        self.extension = (data[0] & self.X_MASK) == self.X_MASK
        cc = data[0] & self.CC_MASK  # pylint:disable=invalid-name
        self.csrc = []  # cc > 0 ? new long[cc] : null
        self.marker = (data[1] & self.M_MASK) == self.M_MASK
        self.payload_type = data[1] & self.PT_MASK
        self.sequence = data[2] * 256 + data[3]
        self.timestamp = ((data[4] * 256 + data[5]) * 256 + data[6]) * 256 + data[7]
        self.ssrc = ((data[8] * 256 + data[9]) * 256 + data[10]) * 256 + data[11]

        for _ in range(cc):
            self.csrc.append(
                ((data[offset] * 256 + data[offset + 1]) * 256 + data[offset + 2]) * 256
                + data[offset + 3])
            offset += 4
            # FIXME In session.c of VLC they store per-source statistics in a rtp_source_t struct

        if self.extension:  # Extension header (ignored for now)
            extension_length = data[offset + 2] * 256 + data[offset + 3]
            offset += 4 + extension_length
            if length < offset:
                self._error = self.ER_EXTENSION_LENGTH
                return

        # And finally ... The payload !
        self.payload = data[offset:length]

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#    public int compareTo(RtpPacket pPacket):
#        BEFORE = -1
#        EQUAL   = 0
#        AFTER   = 1

#        # This optimization is usually worthwhile, and can always be added
#        if (this == pPacket) return EQUAL
#        if (this.sequence < pPacket.sequence) return BEFORE
#        if (this.sequence > pPacket.sequence) return AFTER
#        return EQUAL

    @classmethod
    def create(cls, sequence, timestamp, payload_type, payload):
        """
        Create a valid RTP packet with a given payload.

        :param sequence: Packet sequence number (16 bits)
        :type sequence: int
        :param timestamp: Packet timestamp (32 bits)
        :type timestamp: int
        :param payload_type: Packet payload type (7 bits)
        :type payload_type: int
        :param payload: Packet payload, can be an array of bytes or a string
        :type payload: bytearray

        **Example usage**

        >>> p = RtpPacket.create(10, 1024, RtpPacket.MP2T_PT, 'The payload string')
        >>> q = RtpPacket.create(11, 1028, RtpPacket.MP2T_PT, bytearray.fromhex('00 11 22 33'))
        >>> r = RtpPacket.create(11, 1028, RtpPacket.DYNAMIC_PT, bytearray.fromhex('cc aa ff ee'))
        >>> assert p.validMP2T and q.validMP2T and r.valid
        """
        rtp = cls(None, 0)
        rtp.version = 2
        rtp.padding = False
        rtp.extension = False
        rtp.marker = False
        rtp.payload_type = payload_type & cls.PT_MASK
        rtp.sequence = sequence & cls.S_MASK
        rtp.timestamp = timestamp & cls.TS_MASK
        rtp.ssrc = 0
        rtp.csrc = []
        rtp.payload = payload
        return rtp

    def __eq__(self, other):
        """
        Equality test.

        .. warning::

            Test equality only on some fields (not all) !
        """
        return (
            isinstance(other, self.__class__)
            and self.sequence == other.sequence
            and self.timestamp == other.timestamp
            and self.payload_type == other.payload_type
            and self.payload == other.payload)

    def __str__(self):
        """
        Returns a string containing a formated representation of the packet fields.

        **Example usage**

        >>> rtp = RtpPacket.create(6, 777, RtpPacket.MP2T_PT, 'salut les loulous')
        >>> print(rtp)
        version      = 2
        errors       = []
        padding      = False
        extension    = False
        marker       = False
        payload type = 33
        sequence     = 6
        timestamp    = 777
        clock rate   = 90000
        time         = 0
        ssrc         = 0
        csrc count   = 0
        payload size = 17
        """
        return f"""version      = {self.version}
errors       = {self.errors}
padding      = {self.padding}
extension    = {self.extension}
marker       = {self.marker}
payload type = {self.payload_type}
sequence     = {self.sequence}
timestamp    = {self.timestamp}
clock rate   = {self.clock_rate}
time         = {self.time:.0f}
ssrc         = {self.ssrc}
csrc count   = {len(self.csrc)}
payload size = {self.payload_size}"""



# pylint:disable=too-many-lines

import argparse, json, os, socket, random, subprocess, sys, time, uuid, yaml

from . import console, filesystem, module, subprocess as py_subprocess  # pylint:disable=reimported
from .argparse import FullPaths, is_dir

_all = module.All(globals())

CONFIG_FILENAME = 'config.yaml'
METADATA_FILENAME = 'metadata.yaml'

DEFAULT_ENVIRONMENTS_FILE = os.path.abspath(os.path.expanduser('~/.juju/environments.yaml'))
DEFAULT_OS_ENV = {
    'APT_LISTCHANGES_FRONTEND': 'none',
    'CHARM_DIR': '/var/lib/juju/units/oscied-storage-0/charm',
    'DEBIAN_FRONTEND': 'noninteractive',
    '_JUJU_CHARM_FORMAT': '1',
    'JUJU_AGENT_SOCKET': '/var/lib/juju/units/oscied-storage-0/.juju.hookcli.sock',
    'JUJU_CLIENT_ID': 'constant',
    'JUJU_ENV_UUID': '878ca8f623174911960f6fbed84f7bdd',
    'JUJU_PYTHONPATH': ':/usr/lib/python2.7/dist-packages:/usr/lib/python2.7'
                       ':/usr/lib/python2.7/plat-x86_64-linux-gnu'
                       ':/usr/lib/python2.7/lib-tk'
                       ':/usr/lib/python2.7/lib-old'
                       ':/usr/lib/python2.7/lib-dynload'
                       ':/usr/local/lib/python2.7/dist-packages'
                       ':/usr/lib/pymodules/python2.7',
    '_': '/usr/bin/python',
    'JUJU_UNIT_NAME': 'oscied-storage/0',
    'PATH': '/usr/local/sbin:/usr/local/bin:/usr/bin:/usr/sbin:/sbin:/bin',
    'PWD': '/var/lib/juju/units/oscied-storage-0/charm',
    'SHLVL': '1'
}

ALL_STATES = PENDING, INSTALLED, STARTED, STOPPED, NOT_STARTED, ERROR, UNKNOWN = \
    ('pending', 'installed', 'started', 'stopped', 'not-started', 'error', 'unknown')

UNKNOWN_STATES = (UNKNOWN,)
PENDING_STATES = (PENDING, INSTALLED)
STARTED_STATES = (STARTED,)
STOPPED_STATES = (STOPPED,)
ERROR_STATES = (ERROR, NOT_STARTED)

# Some Amazon EC2 constraints, waiting for instance-type to be implemented ...
M1_SMALL = 'arch=amd64 cpu-cores=1 cpu-power=100 mem=1.5G'
M1_MEDIUM = 'arch=amd64 cpu-cores=1 cpu-power=200 mem=3.5G'
C1_MEDIUM = 'arch=amd64 cpu-cores=2 cpu-power=500 mem=1.5G'

ENVIRONMENT_COMMANDS = ('destroy-environment', )
SUPER_COMMANDS = ('destroy-environment', )


def juju_do(command, environment=None, options=None, fail=True, log=None, **kwargs):
    """
    Execute a command `command` into environment `environment`.

    **Known issue**:

    Locking Juju status 'are you sure you want to continue connecting (yes/no)'.

    We need a way to confirm our choice
    ``cmd('juju status --environment %s' % environment, cli_input='yes\\n')``
    seem to not work as expected. This happens the first time (and only the first time)
    juju connect to a freshly deployed environment.

    Solution : http://askubuntu.com/questions/123072/ssh-automatically-accept-keys::

        $ echo 'StrictHostKeyChecking no' >> ~/.ssh/config
    """
    is_destroy = (command == 'destroy-environment')
    arguments = ['sudo', 'juju', command] if command in SUPER_COMMANDS else ['juju', command]

    if isinstance(environment, str) and environment != 'default':
        arguments += \
            [environment] if command in ENVIRONMENT_COMMANDS else ['--environment', environment]

    if isinstance(options, list):
        arguments += [str(o) for o in options]

    env = os.environ.copy()
    env['HOME'] = os.path.expanduser('~/')
    env['JUJU_HOME'] = os.path.expanduser('~/.juju')

    if is_destroy:
        # TODO Automate yes answer to destroy-environment
        method = subprocess.check_call if fail else subprocess.call
        return method(arguments)
    result = py_subprocess.cmd(arguments, fail=False, log=log, env=env, **kwargs)

    if result['returncode'] != 0 and fail:
        raise RuntimeError(f"Subprocess failed {' '.join(arguments)} : {result['stderr']}.")

    return yaml.safe_load(result['stdout'])


def load_unit_config(config, log=None):
    """
    Returns a dictionary containing the options names as keys and options default values as values.

    The parameter `config` can be:

    * The path of a charm configuration file (e.g. `config.yaml`).
    * A dictionary containing already loaded options names as keys and options values as values.
    """
    if isinstance(config, str):

        if hasattr(log, '__call__'):
            log(f'Load config from file {config}')

        with open(config, encoding='utf-8') as f:
            options = yaml.safe_load(f)['options']
            config = {}
            for option in options:
                config[option] = options[option]['default']

    for option, value in config.items():
        if str(value).lower() in ('false', 'true'):
            config[option] = str(value).lower() == 'true'
            if hasattr(log, '__call__'):
                log(f'Convert boolean option {option} {value} -> {config[option]}')

    return config


def save_unit_config(path, service, config):
    with open(path, 'w', encoding='utf-8') as f:
        for option, value in config.items():
            if isinstance(value, bool):
                config[option] = str(value)
        config = {service: config}
        f.write(yaml.safe_dump(config))


# Environments -------------------------------------------------------------------------------------

def add_environment(
    environment,
    kind,
    region,
    access_key,
    secret_key,
    control_bucket,
    default_series,
    environments=None
):
    with open(environments or DEFAULT_ENVIRONMENTS_FILE, encoding='utf-8') as f:
        environments_dict = yaml.safe_load(f)

    if environment == 'default':
        raise ValueError(f'Cannot create an environment with name {environment}.')

    if environment in environments_dict['environments']:
        raise ValueError(f'The name {environment} is already used by another environment.')

    if kind == 'ec2':
        environment_dict = {
            'type': kind,
            'region': region,
            'access-key': access_key,
            'secret-key': secret_key,
            'control-bucket': control_bucket,
            'default-series': default_series,
            'ssl-hostname-verification': True,
            'juju-origin': 'ppa',
            'admin-secret': uuid.uuid4().hex
        }
    else:
        raise NotImplementedError(f'Registration of {kind} type of environment.')

    environments_dict['environments'][environment] = environment_dict

    with open(environments, 'w', encoding='utf-8') as f:
        f.write(yaml.safe_dump(environments_dict))

    try:
        return juju_do('bootstrap', environment)
    except RuntimeError as e:
        if 'configuration error' in str(e):
            del environments_dict['environments'][environment]
            with open(environments, 'w', encoding='utf-8') as f:
                f.write(yaml.safe_dump(environments_dict))
            raise ValueError(f'Cannot add environment {environment} ({e}).') from e
        raise


def get_environment(environment, environments=None, get_status=False, status_timeout=None):
    with open(environments or DEFAULT_ENVIRONMENTS_FILE, encoding='utf-8') as f:
        environments_dict = yaml.safe_load(f)

    environment = environments_dict['default'] if environment == 'default' else environment
    try:
        environment_dict = environments_dict['environments'][environment]
    except KeyError as e:
        raise ValueError(f'No environment with name {environment}.') from e

    if get_status:
        environment_dict['status'] = Environment(name=environment).status(timeout=status_timeout)

    return environment_dict


def get_environments(environments=None, get_status=False, status_timeout=None):
    with open(environments or DEFAULT_ENVIRONMENTS_FILE, encoding='utf-8') as f:
        environments_dict = yaml.safe_load(f)

    environments = {}
    for environment in environments_dict['environments'].items():
        information = environment[1]
        if get_status:
            try:
                information['status'] = \
                    Environment(name=environment[0]).status(timeout=status_timeout)
            except RuntimeError:
                information['status'] = UNKNOWN
        environments[environment[0]] = information

    return environments, environments_dict['default']


def get_environments_count(environments=None):
    with open(environments or DEFAULT_ENVIRONMENTS_FILE, encoding='utf-8') as f:
        return len(yaml.safe_load(f)['environments'])


# Units --------------------------------------------------------------------------------------------

def get_unit_path(service, number, *args):
    return os.path.join(f'/var/lib/juju/agents/unit-{service}-{number}/charm', *args)


# Helpers ------------------------------------------------------------------------------------------

__get_ip = None  # pylint:disable=invalid-name


def get_ip():
    global __get_ip  # pylint:disable=global-statement,invalid-name
    if __get_ip is None:
        host = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        host.connect(('8.8.8.8', 80))
        __get_ip = host.getsockname()[0]
        host.close()
    return __get_ip


class CharmConfig(object):

    def __init__(self):
        self.verbose = False

    def __repr__(self):
        return str(self.__dict__)


class CharmHooks(object):  # pylint:disable=too-many-instance-attributes,too-many-public-methods
    """
    A base class to build charms based on python hooks, callable even if juju is not installed.

    Attribute `local_config` must be set by derived class to an instance of
    `serialization.PickelableObject`. This should be loaded from a file that is local to the unit
    by `__init__`. This file is used to store service/ unit-specific configuration. In EBU's project
    called OSCIED, this file is even read by the encapsulated python code of the worker
    (celery daemon).

    **Example usage**

    >>> class MyCharmHooks(CharmHooks):
    ...
    ...     def hook_install(self):
    ...         self.debug('hello world, install some packages with self.cmd(...)')
    ...
    ...     def hook_config_changed(self):
    ...         self.remark('update services based on self.config and update self.local_config')
    ...
    ...     def hook_start(self):
    ...         self.info('start services')
    ...
    ...     def hook_stop(self):
    ...         self.info('stop services')
    ...
    >>> import os
    >>> here = os.path.abspath(os.path.expanduser(os.path.dirname(__file__)))
    >>> here = os.path.join(here, '../../..' if 'build/lib' in here else '..', 'tests')
    >>> metadata = os.path.join(here, 'metadata.yaml')
    >>> config = os.path.join(here, 'config.yaml')

    Trigger some hooks:

    >>> my_hooks = MyCharmHooks(metadata, config, DEFAULT_OS_ENV, force_disable_juju=True)
    [DEBUG] Using juju False, reason: Disabled by user.
    [DEBUG] Load metadata from file ...
    >>>
    >>> my_hooks.trigger('install')
    [HOOK] Execute MyCharmHooks hook install
    [DEBUG] hello world, install some packages with self.cmd(...)
    [HOOK] Exiting MyCharmHooks hook install
    >>>
    >>> my_hooks.trigger('config-changed')
    [HOOK] Execute MyCharmHooks hook config-changed
    [REMARK] update services based on self.config and update self.local_config !
    [HOOK] Exiting MyCharmHooks hook config-changed
    >>>
    >>> my_hooks.trigger('start')
    [HOOK] Execute MyCharmHooks hook start
    [INFO] start services
    [HOOK] Exiting MyCharmHooks hook start
    >>>
    >>> my_hooks.trigger('stop')
    [HOOK] Execute MyCharmHooks hook stop
    [INFO] stop services
    [HOOK] Exiting MyCharmHooks hook stop
    >>>
    >>> my_hooks.trigger('not_exist')
    Traceback (most recent call last):
        ...
    AttributeError: 'MyCharmHooks' object has no attribute 'hook_not_exist'
    """

    def __init__(self, metadata, default_config, default_os_env, force_disable_juju=False):
        self.config = CharmConfig()
        self.local_config = None
        reason = 'Life is good !'
        try:
            if force_disable_juju:
                raise OSError('Disabled by user.')
            self.juju_ok = True
            self.load_config(json.loads(self.cmd(['config-get', '--format=json'])['stdout']))
            self.env_uuid = os.environ.get('JUJU_ENV_UUID')
            self.name = os.environ['JUJU_UNIT_NAME']
            self.private_address = socket.getfqdn(self.unit_get('private-address'))
            self.public_address = socket.getfqdn(self.unit_get('public-address'))
        except (subprocess.CalledProcessError, OSError) as e:
            reason = e
            self.juju_ok = False
            if default_config is not None:
                self.load_config(default_config)
            self.env_uuid = default_os_env['JUJU_ENV_UUID']
            self.name = default_os_env['JUJU_UNIT_NAME']
            self.private_address = self.public_address = socket.getfqdn(get_ip())
        self.directory = os.getcwd()
        self.debug(f'Using juju {self.juju_ok}, reason: {reason}')
        self.load_metadata(metadata)

    # ----------------------------------------------------------------------------------------------

    @property
    def identifier(self):
        """
        Returns the id extracted from the unit's name.

        **Example usage**

        >>> hooks = CharmHooks(None, None, DEFAULT_OS_ENV, force_disable_juju=True)
        >>> hooks.name = 'oscied-storage/3'
        >>> hooks.identifier
        3
        """
        return int(self.name.split('/')[1])

    @property
    def name_slug(self):
        return self.name.replace('/', '-')

    # TODO add cache decorator
    @property
    def is_leader(self):
        """
        Returns True if current unit is the leader of the peer-relation.

        By convention, the leader is the unit with the *smallest* id (the *oldest* unit).
        """
        try:
            rel_ids = self.relation_ids('peer')
            if not rel_ids:
                return True  # no peer relation, so we're a leader that feels alone !
            assert len(rel_ids) == 1, f'Expect only 1 peer relation id: {rel_ids}'
            peers = self.relation_list(rel_ids[0])
            self.debug(f'us={self.name} peers={peers}')
            return len(peers) == 0 or self.name <= min(peers)
        except Exception as e:  # pylint:disable=broad-except
            self.remark(f'Bug during leader detection: {repr(e)}')
            return True

    # Maps calls to charm helpers functions and replace them if called in standalone ---------------

    def log(self, message):
        if self.juju_ok:
            return self.cmd(['juju-log', message], logging=False)  # Avoid infinite loop !
        print(message)
        return None

    def open_port(self, port, protocol='TCP'):
        if self.juju_ok:
            return self.cmd(['open-port', f'{port}/{protocol}'])
        return self.debug(f'Open port {port} ({protocol})')

    def close_port(self, port, protocol='TCP'):
        if self.juju_ok:
            return self.cmd(['close-port', f'{port}/{protocol}'])
        return self.debug(f'Close port {port} ({protocol})')

    # TODO add memoize decorator
    def unit_get(self, attribute):
        if self.juju_ok:
            return self.cmd(['unit-get', attribute])['stdout'].strip()
        raise NotImplementedError('FIXME juju-less unit_get not yet implemented')

    # TODO add memoize decorator
    def relation_get(self, attribute=None, unit=None, relation_id=None):
        if self.juju_ok:
            command = ['relation-get']
            if relation_id is not None:
                command += ['-r', relation_id]
            command += [_f for _f in (attribute, unit) if _f]
            return self.cmd(command)['stdout'].strip()
        raise NotImplementedError('FIXME juju-less relation_get not yet implemented')

    # TODO add memoize decorator
    def relation_ids(self, relation_name=''):
        if self.juju_ok:
            result = self.cmd(['relation-ids', '--format', 'json', relation_name], fail=False)
            return json.loads(result['stdout']) if result['returncode'] == 0 else None
        raise NotImplementedError('FIXME juju-less relation_ids not yet implemented')

    # TODO add memoize decorator
    def relation_list(self, relation_id=None):
        if self.juju_ok:
            command = ['relation-list', '--format', 'json']
            if relation_id is not None:
                command += ['-r', relation_id]
            result = self.cmd(command, fail=False)
            return json.loads(result['stdout']) if result['returncode'] == 0 else None
        raise NotImplementedError('FIXME juju-less relation_list not yet implemented')

    def relation_set(self, **kwargs):
        if self.juju_ok:
            command = ['relation-set']
            command += [f'{k}={v}' for k, v in kwargs.items()]
            return self.cmd(command)
        raise NotImplementedError('FIXME juju-less relation_set not yet implemented')

    # Convenience methods for logging --------------------------------------------------------------

    def debug(self, message):
        """Convenience method for logging a debug-related message."""
        return self.log(f'[DEBUG] {message}') if self.config.verbose else None

    def info(self, message):
        """Convenience method for logging a standard message."""
        return self.log(f'[INFO] {message}')

    def hook(self, message):
        """Convenience method for logging the triggering of a hook."""
        return self.log(f'[HOOK] {message}')

    def remark(self, message):
        """Convenience method for logging an important remark."""
        return self.log(f'[REMARK] {message} !')

    # ----------------------------------------------------------------------------------------------

    def load_config(self, config):
        """
        Updates `config` attribute with given configuration.

        **Example usage**

        >>> import os
        >>> here = os.path.abspath(os.path.expanduser(os.path.dirname(__file__)))
        >>> config = os.path.join(
        ...     here, '../../..' if 'build/lib' in here else '..', 'tests/config.yaml')
        >>>
        >>> hooks = CharmHooks(None, None, DEFAULT_OS_ENV, force_disable_juju=True)
        >>> hasattr(hooks.config, 'pingu') or hasattr(hooks.config, 'rabbit_password')
        False
        >>> hooks.load_config({'pingu': 'bi bi'})
        >>> hooks.config.pingu
        'bi bi'
        >>> hooks.config.verbose = True
        >>> hooks.load_config(config)
        [DEBUG] Load config from file ...
        [DEBUG] Convert boolean option ... true -> True
        [DEBUG] Convert boolean option ... true -> True
        [DEBUG] Convert boolean option ... true -> True
        >>> hasattr(hooks.config, 'rabbit_password')
        True
        """
        self.config.__dict__.update(load_unit_config(config, log=self.debug))

    def save_local_config(self):
        """
        Save or update local configuration file only if this instance has the attribute
        `local_config`.
        """
        if self.local_config is not None:
            self.debug(f'Save (updated) local configuration {self.local_config}')
            self.local_config.write()

    def load_metadata(self, metadata):
        """
        Set `metadata` attribute with given metadata, `metadata` can be:

        * The path of a charm metadata file (e.g. `metadata.yaml`)
        * A dictionary containing the metadata.

        **Example usage**

        >>> import os
        >>> here = os.path.abspath(os.path.expanduser(os.path.dirname(__file__)))
        >>> metadata = os.path.join(
        ...     here, '../../..' if 'build/lib' in here else '..', 'tests/metadata.yaml')
        >>> hooks = CharmHooks(None, None, DEFAULT_OS_ENV, force_disable_juju=True)
        >>> hooks.metadata
        >>> hooks.load_metadata({'ensemble': 'oscied'})
        >>> hooks.metadata
        {'ensemble': 'oscied'}
        >>> hooks.config.verbose = True
        >>> hooks.load_metadata(metadata)
        [DEBUG] Load metadata from file ...
        >>> hooks.metadata['maintainer']
        'OSCIED Main Developper <david.fischer.ch@gmail.com>'
        """
        if isinstance(metadata, str):
            self.debug(f'Load metadata from file {metadata}')
            with open(metadata, encoding='utf-8') as f:
                metadata = yaml.safe_load(f)
        self.metadata = metadata

    # ----------------------------------------------------------------------------------------------

    def cmd(self, command, logging=True, **kwargs):
        """
        Calls the `command` and returns a dictionary with `stdout`, `stderr`, and the `returncode`.
        """
        return py_subprocess.cmd(command, log=self.debug if logging else None, **kwargs)

    def template_to_config(self, template, config, values):
        filesystem.from_template(template, config, values)
        self.remark(f'File {config} successfully generated')

    # ----------------------------------------------------------------------------------------------

    def trigger(self, hook_name=None):
        """
        Triggers a hook specified in `hook_name`, defaults to ``sys.argv[1]``.

        Hook's name is the nice hook name that one can find in official juju documentation.
        For example if `config-changed` is mapped to a call to ``self.hook_config_changed()``.

        A `ValueError` containing a usage string is raised if a bad number of argument is given.
        """
        if hook_name is None:
            if len(sys.argv) != 2:
                raise ValueError(f'Usage {sys.argv[0]} hook_name (e.g. config-changed)')
            hook_name = sys.argv[1]

        try:  # Call the function hooks_...
            self.hook(f'Execute {self.__class__.__name__} hook {hook_name}')
            getattr(self, f"hook_{hook_name.replace('-', '_')}")()
            self.save_local_config()
        except subprocess.CalledProcessError as e:
            self.log('Exception caught:')
            self.log(e.output)
            raise
        finally:
            self.hook(f'Exiting {self.__class__.__name__} hook {hook_name}')


class Environment(object):  # pylint:disable=too-many-instance-attributes,too-many-public-methods

    def __init__(
        self,
        name='default',
        charms_path='charms',
        config=CONFIG_FILENAME,
        release=None,
        auto=False,
        min_timeout=15
    ):
        self.name = name
        self.charms_path = os.path.abspath(charms_path)
        self.config = config
        self.release = release
        self.auto = auto
        self.min_timeout = min_timeout
        self._properties = None

    def properties(self, cached=False):
        """
        Return a dictionary with properties of the environments as returned by juju get-environment
        (type, ...).

        Set cached to True to allow using cached value of "properties". Caching should be used only
        if the properties you are interested to read, e.g. type, does not change at all!
        """
        if not self._properties or not cached:
            self._properties = juju_do('get-environment', self.name)
        return self._properties

    def status(self, fail=False, timeout=15):
        """
        Return the status of the environment or None in case of time-out
        [TODO verify other conditions].
        """
        if timeout is not None and timeout < self.min_timeout:
            raise ValueError(
                f'A time-out of {timeout} for status is dangerous, '
                f'please increase it to {self.min_timeout}+ or set it to None')
        status_dict = juju_do('status', self.name, timeout=timeout, fail=fail)
        if fail and not status_dict:
            raise RuntimeError(f'Unable to retrieve status of environment {self.name}.')
        return status_dict

    def is_bootstrapped(self, timeout=15):
        """Return True if the environment is bootstrapped (status returns something)."""
        try:
            return bool(self.status(timeout=timeout))
        except Exception:  # pylint:disable=broad-except
            return False

    def symlink_local_charms(self, default_path='default'):
        """Symlink charms default directory to directory of current release."""
        release_symlink = os.path.join(self.charms_path, self.release)
        filesystem.remove(release_symlink)
        filesystem.symlink(os.path.join(self.charms_path, default_path), release_symlink)

    def sync_tools(self, all_tools=True):
        """Copy tools from the official bucket into a local environment."""
        options = ['--all'] if all_tools else None
        return juju_do('sync-tools', self.name, options=options)

    def bootstrap(
        self,
        cleanup=True,
        synchronize_tools=False,
        wait_started=False,
        started_states=STARTED_STATES,
        error_states=ERROR_STATES,
        timeout=600,
        status_timeout=15,
        polling_delay=30
    ):  # pylint:disable=too-many-branches,too-many-locals
        """
        Bootstrap an environment, (optional) terminate all machines and other associated resources
        before the bootstrap.

        :param kwargs: Extra arguments for juju_do(), options is not allowed.
        """
        if cleanup:
            print(f'Cleanup and bootstrap environment {self.name}')
            print('[WARNING] This will terminate all units deployed into environment '
                  f'{self.name} by juju !')
        else:
            print(f'Bootstrap environment {self.name}')

        if not self.auto and not console.confirm('do it now', default=False):
            return None

        if cleanup:
            self.destroy(force=True, remove_default=True)
        if synchronize_tools:
            self.sync_tools(all_tools=True)
        try:
            result = juju_do('bootstrap', self.name)
        except RuntimeError as e:
            result = None
            if 'already' not in str(e):
                raise

        if not wait_started:
            return result

        start_time = time.time()
        while True:
            time_zero = time.time()
            try:
                state = self.status(timeout=status_timeout)['machines']['0']['agent-state']
            except (KeyError, TypeError):
                state = UNKNOWN

            delta_time = time.time() - start_time
            timeout_time = timeout - delta_time

            _msg = f' in {timeout_time:.0f} seconds' if timeout_time > 0 else '!'
            print(f'State of juju bootstrap machine is {state}, time-out{_msg}')

            if state in started_states:
                print(f'Environment bootstrapped in approximatively {delta_time:.0f} seconds')
                break
            if state in error_states:
                raise RuntimeError(f'Bootstrap failed with state {state}.')
            if delta_time > timeout:
                raise TimeoutError(f'Bootstrap time-out with state {state}.')
            time.sleep(max(0, polling_delay - (time.time() - time_zero)))
        return result

    def destroy(
        self,
        environments=None,
        force=True,
        remove_default=False,
        remove=False,
        timeout=15
    ):
        # TODO simpler algorithm
        with open(environments or DEFAULT_ENVIRONMENTS_FILE, encoding='utf-8') as f:
            environments_dict = yaml.safe_load(f)

        name = environments_dict['default'] if self.name == 'default' else self.name

        if name not in environments_dict['environments']:
            raise IndexError(f'No environment with name {name}.')

        if not remove_default and name == environments_dict['default']:
            raise RuntimeError(f'Cannot remove default environment {name}.')

        options = ['--force'] if force else []
        result = None

        if self.is_bootstrapped(timeout=timeout):
            result = juju_do('destroy-environment', name, options)

        if remove:
            # Check if environment destroyed otherwise a lot of trouble with $/€ !
            if self.is_bootstrapped(timeout=timeout):
                raise RuntimeError(f'Environment {name} not removed, it is still alive.')

            del environments_dict['environments'][name]
            with open(environments, 'w', encoding='utf-8') as f:
                f.write(yaml.safe_dump(environments_dict))
        return result

    # Services

    def get_service_config(self, service, fail=True):
        return juju_do('get', self.name, options=[service], fail=fail)

    def get_service(self, service, default=None, fail=True, timeout=15):
        status_dict = self.status(fail=fail, timeout=timeout)
        if not status_dict:
            return default
        try:
            return status_dict['services'][service]  # pylint: disable=unsubscriptable-object
        except KeyError as e:
            if fail:
                raise RuntimeError(
                    f'Service {service} not found in environment {self.name}.') from e
        return default

    def expose_service(self, service, fail=True):
        return juju_do('expose', self.name, options=[service], fail=fail)

    def unexpose_service(self, service, fail=True):
        return juju_do('unexpose', self.name, options=[service], fail=fail)

    def destroy_service(self, service, fail=True):
        return juju_do('destroy-service', self.name, options=[service], fail=fail)

    # Units

    def ensure_num_units(
        self,
        charm,
        service,
        constraints=None,
        expose=False,
        local=True,
        num_units=1,
        release=None,
        repository=None,
        required=True,
        terminate=False,
        to=None,
        units_number_to_keep=None,
        timeout=None
    ):  # pylint:disable=invalid-name,too-many-branches,too-many-locals
        """
        Ensure `num_units` units of `service` into `environment` by adding new or destroying useless
        units first !

        At the end of this method, the number of running units can be greater as `num_units` because
        this algorithm will not destroy units with number in `units_number_to_keep`.

        Some of the argument are forwarded to underlying methods (add_units, deploy_units,
        destroy_unit or nothing) depending on the required action.

        The scale-down algorithm will sort the units by status to kill the useless units before any
        others!

        * Set local to False to deploy non-local charms.
        * Set num_units to None to ensure the service does not exist.
        * Set required to False to bypass this method in automatic mode.
          WARNING: IT MEANS NO ACTION AT ALL.
        * Set units_number_to_keep to a (list, tuple, ...) with the units you want to protect
          against destruction.

        **Example usage**

        ::

            >> from functools import partial
            >> e = Environment(config=None)
            >> e.bootstrap(wait_started=True)
            >> ensure = partial(e.ensure_num_units, charm='ubuntu', local=False, terminate=True)

        Ensure the charm ubuntu is deployed as the service vanilla with 10 machines, and exposed::

            >> ensure(service='vanilla', expose=True, num_units=5)
            Deploy ubuntu as vanilla (ensure 5 instances)
            do it now ? [y/N]: y
            {'deploy_units': None, 'expose_service': None}

        Ensure the service unwanted does not exist (was not deployed)::

            >> ensure(service='unwanted', num_units=None)
            {}

        Ensure the service vanilla scale-down to 1 unit, keeping unit number 3
        (previously 5 machines)::

            >> ensure(service='vanilla', num_units=1, units_number_to_keep=[3])
            {'destroy_unit': {0: {'agent-state': 'started',
                                  'agent-version': '1.16.6.1',
                                  'machine': '1',
                                  'public-address': '10.0.3.115'},
                              4: {'agent-state': 'pending',
                                  'agent-version': '1.16.6.1',
                                  'machine': '5',
                                  'public-address': '10.0.3.213'},
                              2: {'agent-state': 'started',
                                  'agent-version': '1.16.6.1',
                                  'machine': '3',
                                  'public-address': '10.0.3.23'},
                              1: {'agent-state': 'started',
                                  'agent-version': '1.16.6.1',
                                  'machine': '2',
                                  'public-address': '10.0.3.170'}}}
            >> ensure(service='vanilla', num_units=1, units_number_to_keep=[3])
            {}

        Scale-up the service vanilla again::

            >> ensure(service='vanilla', num_units=2)
            {'add_units': None}

        Scale-down the service vanilla to 0 units but keep it alive::

            >> ensure(service='vanilla', num_units=0)
            {'destroy_unit': {3: {'agent-state': 'started',
                                   'agent-version': '1.16.6.1',
                                   'machine': '4',
                                   'public-address': '10.0.3.50'},
                               5: {'agent-state': 'started',
                                   'agent-version': '1.16.6.1',
                                   'machine': '6',
                                   'public-address': '10.0.3.129'}}}

        Ensure the service another has 0 units but doesn't allow his destruction
        (the service does not exist)::

            >> ensure(service='another', num_units=0)
            {}

        Ensure the service vanilla does not exist at all::

            >> ensure(service='vanilla', num_units=None)
            {'destroy_service': None}

        """
        results = {}
        release = release or self.release
        repository = repository or (self.charms_path if local else None)
        s = '' if num_units is None or num_units < 2 else 's'

        print(f'Deploy {charm} as {service or charm} (ensure {num_units} instance{s})')

        if self.auto and required or console.confirm('do it now', default=False):

            assert num_units is None or num_units >= 0
            units_dict = self.get_units(service, default=None, fail=False, timeout=timeout)
            units_count = None if units_dict is None else len(units_dict)

            if num_units is None and units_count is not None:

                results['destroy_service'] = self.destroy_service(service)

            elif units_count is None:

                if num_units:  # avoid to deploy units if asked number of units is 0 or None
                    results['deploy_units'] = self.deploy_units(
                        charm, service,
                        num_units=num_units,
                        to=to,
                        config=self.config,
                        constraints=constraints,
                        local=local,
                        release=release,
                        repository=repository
                    )

            elif units_count < num_units:

                num_units = num_units - units_count
                if num_units:  # avoid to add units if number of units to add is 0
                    results['add_units'] = self.add_units(
                        service or charm,
                        num_units=num_units,
                        to=to)

            elif units_count > num_units:

                num_units = units_count - num_units
                destroyed = {}
                units_number_to_keep = \
                    [int(n) for n in units_number_to_keep] if units_number_to_keep else []

                # TODO implement status comparison for sorting ??
                for status in (ERROR, NOT_STARTED, PENDING, INSTALLED, STARTED):
                    if num_units == 0:
                        break
                    for number, unit_dict in list(units_dict.items()):
                        if num_units == 0:
                            break
                        if number in units_number_to_keep:
                            continue
                        unit_status = unit_dict.get('agent-state', status)
                        if unit_status == status or unit_status not in ALL_STATES:
                            self.destroy_unit(service, number, terminate=False)
                            destroyed[number] = unit_dict
                            del units_dict[number]
                            num_units -= 1

                if terminate:
                    time.sleep(5)
                    for unit_dict in destroyed.values():
                        # FIXME handle fail (multiple units same machine, machine busy, missing ...)
                        self.destroy_machine(unit_dict['machine'])
                results['destroy_unit'] = destroyed
                return results

            if expose:
                results['expose_service'] = self.expose_service(service)

        return results

    def destroy_unit(self, service, number, terminate, delay_terminate=5, fail=True, timeout=None):
        name = f'{service}/{number}'
        unit_dict = self.get_unit(service, number, default=None, fail=fail, timeout=timeout)
        if unit_dict is None:
            return None
        if terminate:
            juju_do('destroy-unit', self.name, options=[name])
            # TODO ideally a flag https://bugs.launchpad.net/juju-core/+bug/1206532
            time.sleep(delay_terminate)
            return self.destroy_machine(unit_dict['machine'])
        return juju_do('destroy-unit', self.name, options=[name])

    def get_unit(self, service, number, default=None, fail=True, timeout=None):
        # TODO maybe none if missing or something else
        name = f'{service}/{number}'
        service_dict = self.get_service(service, default=None, fail=fail, timeout=timeout)
        if service_dict is not None:
            try:
                return service_dict['units'][name]
            except KeyError as e:
                if fail:
                    raise RuntimeError(
                        f'No unit with name {name} on environment {self.name}.') from e
        return default

    def get_unit_public_address(self, service, number):
        """
        Return the public address of a unit. Use the most reliable value available
        (dns-name of the machine).
        """
        if self.properties(cached=True)['type'] == 'local':
            return self.get_unit(service, number)['public-address']

        # public-address may report a private address (172.x.x.x) with non-local deployments see
        # OSCIED #132, so we need this workaround (which cannot be used for local deployments).
        machine_number = self.get_unit(service, number)['machine']
        status_dict = self.status()
        if status_dict:
            return status_dict['machines'][machine_number]['dns-name']
        raise ValueError(f'Unable to get public address of unit {service}/{number}')

    def wait_unit(
        self,
        service,
        number,
        started_states=STARTED_STATES,
        error_states=ERROR_STATES,
        timeout=180,
        polling_timeout=15,
        polling_delay=30
    ):
        start_time = time.time()
        while True:
            time_zero = time.time()
            try:
                state = self.get_unit(service, number, timeout=polling_timeout)['agent-state']
            except (KeyError, TypeError):
                state = UNKNOWN
            delta_time = time.time() - start_time
            if state in started_states:
                return state
            if state in error_states:
                raise RuntimeError(f'State of unit {service}/{number} is {state}')
            if delta_time > timeout:
                raise TimeoutError(f'State of unit {service}/{number} is {state}')
            time.sleep(max(0, polling_delay - (time.time() - time_zero)))

    def add_units(self, service, num_units=1, to=None):  # pylint:disable=invalid-name
        options = ['--num-units', num_units]
        if to is not None:
            options += ['--to', to]
        options += [service]
        return juju_do('add-unit', self.name, options=options)

    def deploy_units(
        self,
        charm,
        service=None,
        num_units=1,
        to=None,
        config=None,
        constraints=None,
        local=False,
        release=None,
        repository=None
    ):  # pylint:disable=invalid-name
        service = service or charm
        if not charm:
            raise ValueError('Charm is required.')
        options = ['--num-units', num_units]
        if to is not None:
            options += ['--to', to]
        if config is not None:
            options += ['--config', config]
        if constraints is not None:
            options += ['--constraints', constraints]
        if release is not None:
            charm = f'{release}/{charm}'
        if local:
            charm = f'local:{charm}'
        if repository is not None:
            options += ['--repository', repository]
        options += [_f for _f in (charm, service) if _f]
        return juju_do('deploy', self.name, options=options)

    def get_units(self, service, default=None, fail=True, timeout=None):
        service_dict = self.get_service(service, default=None, fail=fail, timeout=timeout)
        if service_dict is None:
            return default
        units_dict = service_dict.get('units', {})
        return {int(name.split('/')[1]): info for name, info in units_dict.items()}

    def get_units_count(self, service, default=None, fail=True, timeout=None):
        service_dict = self.get_service(service, default=None, fail=fail, timeout=timeout)
        if service_dict is None:
            return default
        units_dict = service_dict.get('units', {})
        return len(units_dict)

    # Machines

    def cleanup_machines(self, fail=True, timeout=None):
        environment_dict = self.status(fail=fail, timeout=timeout) or {}
        machines = environment_dict.get('machines', {}).keys()
        busy = ['0']  # the machine running the juju daemon !
        for s_dict in environment_dict.get('services', {}).values():
            busy.extend(u_dict.get('machine', None) for u_dict in s_dict.get('units', {}).values())
        idle = [m for m in machines if m not in busy]
        return juju_do('destroy-machine', self.name, options=idle) if idle else None

    def destroy_machine(self, machine):
        return juju_do('destroy-machine', self.name, options=[machine])

    # Relations

    def add_relation(self, service1, service2, relation1=None, relation2=None):
        """Add a relation between 2 services. Knowing that the relation may already exists."""
        print(f'Add relation between {service1} and {service2}')
        if not self.auto and not console.confirm('do it now', default=False):
            return None

        member1 = service1 if relation1 is None else f'{service1}:{relation1}'
        member2 = service2 if relation2 is None else f'{service2}:{relation2}'
        result = None
        try:
            result = juju_do('add-relation', self.name, options=[member1, member2])
        except RuntimeError as e:
            # TODO get status of service before adding relation may be cleaner.
            if 'already exists' not in str(e):
                raise
        return result

    def remove_relation(self, service1, service2, relation1=None, relation2=None):
        """Remove a relation between 2 services. Knowing that the relation may not exists."""
        print(f'Remove relation between {service1} and {service2}')
        if not self.auto and not console.confirm('do it now', default=False):
            return None

        member1 = service1 if relation1 is None else f'{service1}:{relation1}'
        member2 = service2 if relation2 is None else f'{service2}:{relation2}'
        try:
            result = juju_do('remove-relation', self.name, options=[member1, member2])
        except RuntimeError as e:
            # TODO get status of service before removing relation may be cleaner.
            if 'exists' not in str(e):
                raise
        return result


class DeploymentScenario(object):

    def __init__(self, environments, args=None, namespace=None, **kwargs):
        parser = self.get_parser(**kwargs)
        self.args = parser.parse_args(args=args, namespace=namespace)
        self.environments = environments
        for environment in environments:
            environment.auto = self.args.auto
            environment.charms_path = self.args.charms_path
            environment.release = environment.release or self.args.release
            self.__dict__.update({environment.name: environment})  # A shortcut

    @staticmethod
    def get_parser(
        epilog='',
        charms_path='charms',
        release='raring',
        auto=False
    ):  # pylint:disable=invalid-name
        HELP_A = 'Toggle automatic confirmation of the actions, WARNING: Use it with care.'
        HELP_M = 'Directory (repository) of any local charm.'
        HELP_R = 'Ubuntu serie to deploy by JuJu.'
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            epilog=epilog)
        arg = parser.add_argument
        arg('-a', '--auto', action='store_true', default=auto, help=HELP_A)
        arg('-m', '--charms_path', action=FullPaths, default=charms_path, type=is_dir, help=HELP_M)
        arg('-r', '--release', action='store', default=release, help=HELP_R)
        return parser

    def run(self):
        raise NotImplementedError('Here should be implemented the deployment scenario.')


# Simulation ---------------------------------------------------------------------------------------

# DISCLAIMER: Ideally this module will implement a simulated juju_do to make it possible to use the
# same methods to drive a real or a simulated juju process ... The following code is a partial/light
# implementation of ... Do not use it for your own purposes !!!

class SimulatedUnit(object):
    """A simulated unit with a really simple state machine having a latency to start and stop."""

    def __init__(self, start_latency_range, stop_latency_range, state=PENDING):
        self.counter = self.next_state = None
        self.state = state
        self.start_latency_range = start_latency_range
        self.stop_latency_range = stop_latency_range

    def start(self):
        self.counter = random.randint(*self.start_latency_range)
        self.next_state = STARTED

    def stop(self):
        self.counter = random.randint(*self.stop_latency_range)
        self.next_state = STOPPED

    def tick(self):
        if self.counter:
            self.counter -= 1
            if self.counter == 0:
                self.state = self.next_state
                self.next_state = None


class SimulatedUnits(object):
    """Manage a set of simulated units."""

    def __init__(self, start_latency_range, stop_latency_range):
        self.start_latency_range = start_latency_range
        self.stop_latency_range = stop_latency_range  # TODO not yet used by this simulator ...
        self.units = {}
        self.number = 0

    def ensure_num_units(self, num_units=1, units_number_to_keep=None):
        """Ensure `num_units` units by adding new units or destroying useless units first !"""
        assert num_units is None or num_units >= 0

        units_count = len(self.units)

        if num_units is None:
            self.units = {}
            return 'Simulate destruction of service.'

        if units_count == num_units:
            return None

        if units_count < num_units:
            num_units = num_units - units_count
            for _ in range(num_units):
                unit = SimulatedUnit(self.start_latency_range, self.stop_latency_range)
                unit.start()
                self.units[self.number] = unit
                self.number += 1
            return f'Simulate deployment of {num_units} units.'

        # units_count > num_units
        num_units = units_count - num_units
        destroyed = {}
        # Sort units by status to kill the useless units before any others !
        # TODO implement status comparison for sorting ??
        for state in (ERROR, NOT_STARTED, PENDING, INSTALLED, STARTED):
            if num_units == 0:
                break
            for number, unit in self.units.items():
                if num_units == 0:
                    break
                if units_number_to_keep is not None and number in units_number_to_keep:
                    continue
                if unit.state == state or unit.state not in ALL_STATES:
                    destroyed[number] = unit
                    unit.stop()
                    num_units -= 1
        return destroyed

    def tick(self):
        """Increment time of 1 tick and remove units that are in STOPPED state."""
        for number, unit in self.units.items():
            unit.tick()
            if unit.state == STOPPED:
                del self.units[number]  # pylint: disable=unnecessary-dict-index-lookup


__all__ = _all.diff(globals())



"""
Decorators for enhancing your models.
"""

try:
    from django.urls import reverse
except ImportError:
    # For Django < 2.0
    from django.core.urlresolvers import reverse

__all__ = ['with_urls']


def with_urls(base_url, *interface_actions, **kwargs):
    base_url = base_url + ':'

    if not interface_actions:
        raise ValueError('At least one action should be defined')

    singleton = kwargs.pop('singleton', False)

    if kwargs:
        raise AttributeError(kwargs)

    def _with_urls(model):
        model.interface_actions = interface_actions

        for action in interface_actions:
            method_name = f'get_{action}_url'

            if hasattr(model, method_name):
                continue

            if action == 'create':
                @classmethod
                def method(cls):
                    return reverse(base_url + 'create')
            else:
                def get_url(action, singleton):
                    if singleton:
                        def _get_url(self):
                            return reverse(base_url + action)
                    else:
                        def _get_url(self):
                            return reverse(base_url + action, args=[str(self.pk)])
                    return _get_url
                method = get_url(action, singleton)

            method.__name__ = method_name
            setattr(model, method_name, method)

        if 'detail' in interface_actions:
            # Add get_absolute_url for Django to be happy
            model.get_absolute_url = model.get_detail_url

        return model
    return _with_urls



import inspect, itertools
from collections import abc

from . import module
from .collections import merge_dicts

_all = module.All(globals())


def get_arguments_names(function):
    """
    Return a list with arguments names.

    >>> from pytoolbox import types
    >>>
    >>> get_arguments_names(get_arguments_names)
    ['function']
    >>>
    >>> def my_func(directory, a=1, *args, b, c=None, **kwargs):
    ...     pass
    ...
    >>> get_arguments_names(my_func)
    ['directory', 'a', 'args', 'b', 'c', 'kwargs']
    >>>
    >>> get_arguments_names(types.get_subclasses)
    ['obj', 'nested']
    >>> get_arguments_names(types.merge_bases_attribute)
    ['cls', 'attr_name', 'init', 'default', 'merge_func']
    """
    return list(inspect.signature(function).parameters.keys())


def get_properties(obj):
    return (
        (n, getattr(obj, n))
        for n, p in inspect.getmembers(obj.__class__, lambda m: isinstance(m, property))
    )


def get_slots(obj):
    """Return a set with the `__slots__` of the `obj` including all parent classes `__slots__`."""
    return set(itertools.chain.from_iterable(
        getattr(cls, '__slots__', ())
        for cls in obj.__class__.__mro__)
    )


def get_subclasses(obj, nested=True):
    """
    Walk the inheritance tree of ``obj``. Yield tuples with (class, subclasses).

    **Example usage**

    >>> class Root(object):
    ...     pass
    ...
    >>> class NodeA(Root):
    ...     pass
    ...
    >>> class NodeB(Root):
    ...     pass
    ...
    >>> class NodeC(NodeA):
    ...     pass
    ...
    >>> class NodeD(NodeA):
    ...     pass
    ...
    >>> class NodeE(NodeD):
    ...     pass
    ...
    >>> [(c.__name__, bool(s)) for c, s in get_subclasses(Root)]
    [('NodeA', True), ('NodeC', False), ('NodeD', True), ('NodeE', False), ('NodeB', False)]
    >>> [(c.__name__, bool(s)) for c, s in get_subclasses(Root, nested=False)]
    [('NodeA', True), ('NodeB', False)]
    >>> [(c.__name__, bool(s)) for c, s in get_subclasses(NodeB)]
    []
    >>> [(c.__name__, bool(s)) for c, s in get_subclasses(NodeD)]
    [('NodeE', False)]
    """
    for subclass in obj.__subclasses__():
        yield subclass, subclass.__subclasses__()
        if nested:
            for sub_subclass in get_subclasses(subclass, nested):
                yield sub_subclass


def isiterable(obj, blacklist=(bytes, str)):
    """
    Return ``True`` if the object is an iterable, but ``False`` for any class in `blacklist`.

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>> for obj in b'binary', 'unicode', 42:
    ...     asserts.false(isiterable(obj), obj)
    >>> for obj in [], (), set(), iter({}.items()):
    ...     asserts.true(isiterable(obj), obj)
    >>> isiterable({}, dict)
    False
    """
    return isinstance(obj, abc.Iterable) and not isinstance(obj, blacklist)


def merge_annotations(cls: type):
    """
    Merge annotations defined in all bases classes (using `__mro__`) into given `cls`.

    Can be used as a decorator.

    **Example usage**

    >>> class Point2D(object):
    ...     x: int
    ...     y: int
    ...
    >>> class Point3D(Point2D):
    ...     z: int
    ...
    >>> class Point4D(Point3D, Point2D):
    ...     w: int
    ...
    >>> @merge_annotations
    ... class Point4X(Point4D):
    ...     x: float
    ...     other: str
    ...
    >>> assert Point2D.__annotations__ == {'x': int, 'y': int}
    >>> assert Point3D.__annotations__ == {'z': int}
    >>> assert Point4D.__annotations__ == {'w': int}
    >>> assert Point4X.__annotations__ == {'x': float, 'y': int, 'z': int, 'w': int, 'other': str}
    """
    cls.__annotations__ = merge_dicts(*[
        getattr(base, '__annotations__', {})
        for base in reversed(cls.__mro__)
    ])
    return cls


def merge_bases_attribute(cls, attr_name, init, default, merge_func=lambda a, b: a + b):
    """
    Merge all values of attribute defined in all bases classes (using `__mro__`).
    Return resulting value. Use default every time a class does not have given attribute.

    Be careful, `merge_func` must be a pure function.
    """
    value = init
    for base in cls.__mro__:
        value = merge_func(value, getattr(base, attr_name, default))
    return value


class DummyObject(object):
    """
    Easy way to generate a dynamic object with the attributes defined at instantiation.

    **Example usage**

    >>> obj = DummyObject(foo=42, bar=None)
    >>> obj.foo
    42
    >>> obj.bar is None
    True
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class EchoObject(object):
    """
    Object that return any missing attribute as an instance of :class:`EchoObject` with the name set
    to the Python expression used to access it. Also implements __getitem__. Some examples are worth
    hundred words...

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>> something = EchoObject('something', language='Python')
    >>> something._name
    'something'
    >>> something.language
    'Python'
    >>> hasattr(something, 'everything')
    True
    >>> type(something.user.email)
    <class 'pytoolbox.types.EchoObject'>
    >>> str(something.user.first_name)
    'something.user.first_name'
    >>> str(something[0][None]['bar'])
    "something[0][None]['bar']"
    >>> str(something[0].node['foo'].x)
    "something[0].node['foo'].x"
    >>> str(something)
    'something'

    You can also define the class for the generated attributes:

    >>> something.attr_class = list
    >>> type(something.cool)
    <class 'list'>

    This class handles sub-classing appropriately:

    >>> class MyEchoObject(EchoObject):
    ...     pass
    >>>
    >>> type(MyEchoObject('name').x.y.z)
    <class 'pytoolbox.types.MyEchoObject'>
    """
    attr_class = None

    def __init__(self, name, **attrs):
        assert '_name' not in attrs
        self.__dict__.update(attrs)
        self._name = name

    def __getattr__(self, name):
        return (self.attr_class or self.__class__)(f'{self._name}.{name}')

    def __getitem__(self, key):
        return (self.attr_class or self.__class__)(f'{self._name}[{repr(key)}]')

    def __str__(self):
        return self._name


class EchoDict(dict):
    """
    Dictionary that return any missing item as an instance of :class:`EchoObject` with the name set
    to the Python expression used to access it. Some examples are worth hundred words...

    **Example usage**

    >>> context = EchoDict('context', language='Python')
    >>> context._name
    'context'
    >>> context['language']
    'Python'
    >>> 'anything' in context
    True
    >>> str(context['user'].first_name)
    "context['user'].first_name"
    >>> str(context[0][None]['bar'])
    "context[0][None]['bar']"
    >>> str(context[0].node['foo'].x)
    "context[0].node['foo'].x"

    You can also define the class for the generated items:

    >>> context.item_class = set
    >>> type(context['jet'])
    <class 'set'>
    """
    item_class = EchoObject

    def __init__(self, name, **items):
        assert '_name' not in items
        super().__init__(**items)
        self._name = name

    def __contains__(self, key):
        """Always return True because missing items are generated."""
        return True

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.item_class(f'{self._name}[{repr(key)}]')


class MissingType(object):

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __bool__(self):
        return False

    def __repr__(self):
        return 'Missing'


Missing = MissingType()

__all__ = _all.diff(globals())



import logging, uuid

from flask import abort, Response
from werkzeug.exceptions import HTTPException

from .private import ObjectId
from .serialization import object_to_json
from .validation import valid_uuid

__all__ = ['STATUS_TO_EXCEPTION', 'check_id', 'json_response', 'map_exceptions']

STATUS_TO_EXCEPTION = {400: TypeError, 404: IndexError, 415: ValueError, 501: NotImplementedError}


def check_id(value):
    if valid_uuid(value, objectid_allowed=False, none_allowed=False):
        return uuid.UUID(value)
    if ObjectId is not None and valid_uuid(value, objectid_allowed=True, none_allowed=False):
        return ObjectId(value)
    raise ValueError(f'Wrong id format {value}')


def json_response(status, value=None, include_properties=False):
    response = Response(
        response=object_to_json({'status': status, 'value': value}, include_properties),
        status=status,
        mimetype='application/json')
    response.status_code = status
    return response


def map_exceptions(exception):
    """
    Maps a standard `exception` into corresponding HTTP exception class.

    **Example usage**

    >>> from werkzeug import exceptions as w_exceptions
    >>> from pytoolbox.unittest import asserts
    >>> asserts.raises(w_exceptions.BadRequest, map_exceptions, TypeError('test'))
    >>> asserts.raises(w_exceptions.NotFound, map_exceptions, IndexError('test'))
    >>> asserts.raises(w_exceptions.NotImplemented, map_exceptions, NotImplementedError('test'))

    Any instance of HTTPException is simply raised without any mapping:

    >>> asserts.raises(w_exceptions.ImATeapot, map_exceptions, w_exceptions.ImATeapot('test'))
    >>> asserts.raises(w_exceptions.NotFound, map_exceptions, w_exceptions.NotFound('test'))

    Convert a JSON response of kind {'status': 200, 'value': '...'}:

    >>> print(map_exceptions({'status': 200, 'value': 'The good value.'}))
    The good value.
    >>> map_exceptions({'status': 415, 'value': 'The value is bad.'})
    Traceback (most recent call last):
        ...
    ValueError: The value is bad.
    """
    if isinstance(exception, Exception):
        try:
            logging.exception(exception)
        except AttributeError:
            logging.exception(repr(exception))
    if isinstance(exception, dict):
        if exception['status'] == 200:
            return exception['value']
        exception_class = STATUS_TO_EXCEPTION.get(exception['status'], Exception)
        raise exception_class(exception['value'])
    if isinstance(exception, HTTPException):
        raise exception
    if isinstance(exception, TypeError):
        abort(400, str(exception))
    elif isinstance(exception, KeyError):
        abort(400, f'Key {exception} not found.')
    elif isinstance(exception, IndexError):
        abort(404, str(exception))
    elif isinstance(exception, ValueError):
        abort(415, str(exception))
    elif isinstance(exception, NotImplementedError):
        abort(501, str(exception))
    else:
        abort(500, f'{exception.__class__.__name__} {repr(exception)} {str(exception)}')
    return None



from urllib.parse import urljoin

from selenium.webdriver.remote.command import Command
from selenium.webdriver.support import ui

from . import common, exceptions, webdrivers

__all__ = ['LiveClient']


class LiveClient(common.FindMixin):

    web_driver_class = webdrivers.Firefox

    def __init__(self, live_server_url, web_driver_class=None):
        self.live_server_url = live_server_url
        if web_driver_class:
            self.web_driver_class = web_driver_class
        self.web_driver = self.web_driver_class()
        self._css_prefix = ''

    @property
    def css_prefix(self):
        return self._css_prefix

    @css_prefix.setter
    def css_prefix(self, value):
        assert not value or not self._css_prefix, self._css_prefix
        self._css_prefix = value

    def find_css(self, css_selector, prefix=True, force_list=False, fail=True):
        """Shortcut to find elements by CSS. Returns either a list or singleton."""
        if prefix and self.css_prefix:
            css_selector = f'{self.css_prefix} {css_selector}'
        return self.web_driver.find_css(css_selector, force_list=force_list, fail=fail)

    def find_xpath(self, xpath, force_list=False, fail=True):
        return self.web_driver.find_xpath(xpath, force_list=force_list, fail=fail)

    def get(self, url, data=None):
        assert data is None
        url = urljoin(self.live_server_url, url) if '://' not in url else url
        response = type('Response', (object, ), self.web_driver.execute(Command.GET, {'url': url}))
        response.status_code = 200 if self.web_driver.current_url == url else 404
        return response

    def quit(self):
        return self.web_driver.quit()

    def set_element(self, name, value, clear=None):
        """Set the properties of an element. Works with both WebElement and Select."""
        element = self.find_name(name)
        if clear is None:
            clear = element.is_displayed()
        if isinstance(element, self.web_driver.web_element_classes['select']):
            return element.select_by_value(value)
        if clear:
            element.clear()
        return element.send_keys(value)

    def submit(self):
        return self.find_css('form button.btn-primary[type="submit"]').click()

    def wait_for_css(self, css_selector='', inverse=False, prefix=True, timeout=5, fail=True):
        try:
            return ui.WebDriverWait(self.web_driver, timeout).until(
                lambda driver: bool(self.find_css(css_selector, prefix, fail=False)) ^ inverse
            )
        except exceptions.TimeoutException:  # pylint:disable=no-member
            if fail:
                raise
        return None

    def wait_for_id(self, element_id, inverse=False, prefix=True, timeout=5, fail=True):
        return self.wait_for_css(f'#{element_id}', inverse, prefix, timeout, fail)

    def wait_for_name(self, element_name, inverse=False, prefix=True, timeout=5, fail=True):
        return self.wait_for_css(f'[name="{element_name}"]', inverse, prefix, timeout, fail)



from . import exceptions

__all__ = ['FindMixin']


class FindMixin(object):

    @staticmethod
    def clean_elements(elements, criteria, force_list=False, fail=True):
        if elements:
            return elements if force_list or len(elements) > 1 else elements[0]
        if fail:
            raise exceptions.NoSuchElementException(criteria)  # pylint:disable=no-member
        return None

    def find_css(self, css_selector, prefix=True, force_list=False, fail=True):
        assert prefix  # Not implemented
        elements = self.find_elements_by_css_selector(css_selector)
        return self.clean_elements(elements, css_selector, force_list, fail)

    def find_id(self, element_id, prefix=True, force_list=False, fail=True):
        return self.find_css(f'#{element_id}', prefix, force_list, fail)

    def find_name(self, element_name, prefix=True, force_list=False, fail=True):
        return self.find_css(f'[name={element_name}]', prefix, force_list, fail)

    def find_xpath(self, xpath, force_list=False, fail=True):
        return self.clean_elements(self.find_elements_by_xpath(xpath), xpath, force_list, fail)



import functools, os, sys, time, urllib.request, urllib.error, urllib.parse

import requests

from pytoolbox import console, crypto, filesystem, module
from pytoolbox.exceptions import BadHTTPResponseCodeError, CorruptedFileError

_all = module.All(globals())


def download(url, path):
    """Read the content of given `url` and save it as a file `path`."""
    with open(path, 'wb') as target:
        with urllib.request.urlopen(url) as source:
            target.write(source.read())


def iter_download_core(url, code=200, chunk_size=102400, **kwargs):
    response = requests.get(url, stream=bool(chunk_size), **kwargs)
    length = response.headers.get('content-length')
    if response.status_code != code:
        raise BadHTTPResponseCodeError(url=url, code=code, r_code=response.status_code)
    if response.status_code == 200:
        if chunk_size:
            position, length = 0, None if length is None else int(length)
            for chunk in response.iter_content(chunk_size):
                position += len(chunk)
                yield position, length, chunk
        else:
            chunk = response.content
            yield len(chunk), len(chunk), chunk


def iter_download_to_file(
    url,
    path,
    code=200,
    chunk_size=102400,
    force=True,
    hash_algorithm=None,
    expected_hash=None,
    **kwargs
):
    position = length = 0
    chunk = file_hash = None
    downloaded = False
    if force or not os.path.exists(path):
        file = None
        try:
            for position, length, chunk in iter_download_core(url, code, chunk_size, **kwargs):
                downloaded = True
                if hash_algorithm:
                    file_hash = file_hash or crypto.new(hash_algorithm)
                    file_hash.update(chunk)
                file_hash_digest = file_hash.hexdigest() if file_hash else None
                yield position, length, chunk, downloaded, file_hash_digest
                file = file or open(path, 'wb')  # pylint: disable=consider-using-with
                file.write(chunk)
        finally:
            if file:
                file.close()
        if file_hash:
            file_hash = file_hash.hexdigest()

    elif hash_algorithm:
        file_hash = crypto.checksum(
            path,
            is_path=True,
            algorithm=hash_algorithm,
            chunk_size=chunk_size)

    if expected_hash and file_hash != expected_hash:
        raise CorruptedFileError(path=path, file_hash=file_hash, expected_hash=expected_hash)

    if not downloaded:
        yield position, length, chunk, downloaded, file_hash


def download_ext(  # pylint:disable=too-many-locals
    url,
    path,
    code=200,
    chunk_size=102400,
    force=True,
    hash_algorithm=None,
    expected_hash=None,
    progress_callback=None,
    **kwargs
):
    """
    Read the content of given `url` and save it as a file `path`, extended version.

    * Set `code` to expected response code.
    * Set `force` to False to avoid downloading the file if it already exists.
    * Set `hash_algorithm` to a string or a method from the :mod:`hashlib`.
    * Set `expected_hash` to the expected hash value, warning: this will force computing the hash of
      the file.
    * Set `kwargs` to any extra argument accepted by ``requests.get()``.

    If `chunk_size` and `hash_algorithm` are both set then the hash algorithm will be called for
    every chunk of data, this may optimize the performances. In any case, if a hash algorithm is
    defined then you will get a hash even if the file is already downloaded.

    **Example usage**

    >>> import hashlib
    >>> from pytoolbox import filesystem
    >>> from pytoolbox.unittest import asserts
    >>>
    >>> _ = filesystem.remove('small.mp4')
    >>>
    >>> url = 'https://pytoolbox.s3-eu-west-1.amazonaws.com/tests/small.mp4'
    >>>
    >>> download_ext(url, 'small.mp4')
    (False, True, None)
    >>> download_ext(url, 'small.mp4', force=False)
    (True, False, None)
    >>>
    >>> download_ext(url, 'small.mp4', force=False, hash_algorithm='md5',
    ...              expected_hash='a3ac7ddabb263c2d00b73e8177d15c8d')
    (True, False, 'a3ac7ddabb263c2d00b73e8177d15c8d')
    >>>
    >>> def progress(start_time, position, length, chunk):
    ...     print('(%d, %d)' % (position, length))
    >>>
    >>> download_ext(url, 'small.mp4', progress_callback=progress)
    (102400, 383631)
    (204800, 383631)
    (307200, 383631)
    (383631, 383631)
    (True, True, None)
    >>>
    >>> with asserts.raises(CorruptedFileError):
    ...     download_ext(
    ...         url,
    ...         'small.mp4',
    ...         hash_algorithm=hashlib.md5,
    ...         expected_hash='efac5df252145c2d07b73e8177d15c8d')
    >>>
    >>> download_ext('http://techslides.com/monkey.mp4', 'monkey.mp4', code=404)
    (False, False, None)
    >>>
    >>> with asserts.raises(BadHTTPResponseCodeError):
    ...     download_ext('http://techslides.com/monkey.mp4', 'monkey.mp4')
    """
    downloaded = False
    exists = os.path.exists(path)
    file_hash = None
    start_time = time.time()
    for position, length, chunk, downloaded, file_hash in iter_download_to_file(
        url, path, code, chunk_size, force, hash_algorithm, expected_hash, **kwargs
    ):
        if progress_callback:
            progress_callback(start_time, position, length, chunk)
    return exists, downloaded, file_hash


def download_ext_multi(
    resources,
    chunk_size=1024 * 1024,
    progress_callback=console.progress_bar,
    progress_stream=sys.stdout,
    progress_template='\r[{counter} of {total}] [{done}{todo}] {name}'
):
    """
    Download resources, showing a progress bar by default.

    Each element should be a `dict` with the url, path and name keys.
    Any extra item is passed to :func:`iter_download_to_file` as extra keyword arguments.
    """
    for counter, resource in enumerate(sorted(resources, key=lambda r: r['name']), 1):
        kwargs = resource.copy()
        start_time = time.time()
        url = kwargs.pop('url')
        path = kwargs.pop('path')
        name = kwargs.pop('name')

        callback = functools.partial(
            progress_callback,
            stream=progress_stream,
            template=progress_template.format(
                counter=counter,
                done='{done}',
                name=name,
                todo='{todo}',
                total=len(resources)))

        if not os.path.exists(path):
            filesystem.makedirs(os.path.dirname(path))
            try:
                for returned in iter_download_to_file(
                    url, path, chunk_size=chunk_size, force=False, **kwargs
                ):
                    callback(start_time, returned[0], returned[1])
            except Exception:
                filesystem.remove(path)
                raise

        callback(start_time, 1, 1)
        progress_stream.write(os.linesep)


__all__ = _all.diff(globals())



"""
Some utilities related to the URLs.
"""

try:
    from django.urls import get_resolver
except ImportError:
    # For Django < 2.0
    from django.core.urlresolvers import get_resolver

from pytoolbox.regex import UUID_REGEX

__all__ = ['INT_PK', 'UUID_PK', 'get_named_patterns']

INT_PK = r'(?P<pk>\d+)'
UUID_PK = r'(?P<pk>%s)' % UUID_REGEX


def get_named_patterns():
    """Returns a generator containing (pattern name, pattern) tuples."""
    return (
        (k, v[0][0][0])
        for k, v in get_resolver(None).reverse_dict.items()
        if isinstance(k, str)
    )



try:
    from bson.objectid import InvalidId, ObjectId  # pylint:disable=unused-import
except ImportError:
    InvalidId = ObjectId = None


def _parse_kwargs_string(kwargs_string, **types):
    if not kwargs_string:
        return {}
    kwargs_list = [kwarg.strip().split('=') for kwarg in kwargs_string.split(';')]
    return {k: types[k](v) for k, v in kwargs_list}



"""
Module related to parsing arguments from the command-line.

**Example usage**

>>> import argparse
>>> from pathlib import Path
>>> parser = argparse.ArgumentParser(
...     formatter_class=argparse.ArgumentDefaultsHelpFormatter,
...     epilog='My super cool software.')
>>> x = parser.add_argument('directory', action=FullPaths, type=is_dir)
>>> str(parser.parse_args(['/usr/lib']).directory)
'/usr/lib'
>>> assert parser.parse_args(['.']).directory == Path(os.getcwd()).expanduser().resolve()
>>> parser.parse_args(['/does_not_exist/'])
Traceback (most recent call last):
    ...
SystemExit: 2
"""

from pathlib import Path
from typing import Optional, Union
import argparse, getpass, os, shutil

from . import itertools, module

_all = module.All(globals())

# Credits https://gist.github.com/brantfaircloth/1443543


def is_dir(path: Union[str, Path]) -> Path:
    """Check if `path` is an actual directory and return it."""
    path = Path(path)
    if path.is_dir():
        return path
    raise argparse.ArgumentTypeError(f'{path} is not a directory')


def is_file(path: Union[str, Path]) -> Path:
    """Check if `path` is an actual file and return it."""
    path = Path(path)
    if path.is_file():
        return path
    raise argparse.ArgumentTypeError(f'{path} is not a file')


def multiple(f):
    """Return a list with the result of `f`(value) for value in values."""
    def _multiple(values):
        return [f(v) for v in values] if isinstance(values, (list, tuple)) else f(values)
    return _multiple


def password(value: Optional[str]) -> str:
    return value or getpass.getpass('Password: ')


def set_columns(value=None, default=120):
    if value is None:
        try:
            value = shutil.get_terminal_size().columns
        except AttributeError:
            value = default
    os.environ['COLUMNS'] = str(value)


class FullPaths(argparse.Action):
    """Expand user/relative paths."""

    def __call__(self, parser, namespace, values, option_string=None):
        def fullpath(path: Union[str, Path]) -> Path:
            return Path(path).expanduser().resolve()
        value = itertools.extract_single([fullpath(v) for v in itertools.chain(values)])
        setattr(namespace, self.dest, value)


class Range(object):

    def __init__(self, type, min, max):  # pylint:disable=redefined-builtin
        self.type = type
        self.min = min
        self.max = max

    def __call__(self, value):
        try:
            value = self.type(value)
        except Exception as e:
            raise argparse.ArgumentTypeError(f'Must be of type {self.type.__name__}') from e
        if not (self.min <= value <= self.max):  # pylint:disable=superfluous-parens
            raise argparse.ArgumentTypeError(f'Must be in range [{self.min}, {self.max}]')
        return value


class HelpArgumentParser(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        set_columns(kwargs.pop('columns', None))
        super().__init__(*args, formatter_class=HelpFormatter, **kwargs)


class HelpFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass


__all__ = _all.diff(globals())



import collections, inspect

from . import module

_all = module.All(globals())


class StateEnumMetaclass(type):

    def __init__(cls, name, bases, cls_dict):
        super().__init__(name, bases, cls_dict)
        if hasattr(cls, 'TRANSITIONS'):
            cls.ALL_STATES = frozenset(cls.TRANSITIONS.keys())
            cls.FINAL_STATES = frozenset(s for s, t in cls.TRANSITIONS.items() if not t)
            inverse_transitions = collections.defaultdict(set)
            for state, transitions in cls.TRANSITIONS.items():
                for transition in transitions:
                    inverse_transitions[transition].add(state)
            cls.INVERSE_TRANSITIONS = {s: frozenset(t) for s, t in inverse_transitions.items()}


class StateEnumMergeMetaclass(StateEnumMetaclass):

    def __init__(cls, name, bases, cls_dict):
        # Retrieve base "state" classes attributes
        m_states, transitions = collections.defaultdict(set), collections.defaultdict(set)
        for base in bases:
            for key, value in (i for i in inspect.getmembers(base) if i[0].endswith('_STATES')):
                m_states[key].update(value)
            for key, values in base.TRANSITIONS.items():
                transitions[key].update(values)
        # Update "state" class attributes
        for key, value in m_states.items():
            setattr(cls, key, frozenset(value))
        for state in transitions.keys():
            setattr(cls, state, state)
        cls.TRANSITIONS = transitions
        super().__init__(name, bases, cls_dict)


class StateEnum(object, metaclass=StateEnumMetaclass):

    @classmethod
    def get(cls, name):
        if name.lower() == name:
            name = name.upper()
            if name in cls.ALL_STATES:
                return name
            return getattr(cls, name + '_STATES', None)
        return None

    @classmethod
    def get_transit_from(cls, state, auto_inverse=False):
        """
        Return a set with the states having a transition to given `state`.

        If `auto_inverse` is set to True then a tuple is returned containing the smallest set from:

        * (States allowed to transit to given `state`, True)
        * (States not allowed to transit to given `state`, False)
        """
        valid = cls.INVERSE_TRANSITIONS[state]
        if not auto_inverse:
            return valid
        not_valid = cls.ALL_STATES - valid
        return (not_valid, False) if len(valid) > len(not_valid) else (valid, True)


__all__ = _all.diff(globals())



"""
Mix-ins for building your own test runners.
"""

import tempfile

from django.conf import settings

__all__ = ['CeleryInMemoryMixin', 'FastPasswordHasherMixin', 'TemporarySendfileRootMixin']


class CeleryInMemoryMixin(object):

    def setup_test_environment(self):
        super().setup_test_environment()
        settings.BROKER_BACKEND = 'memory'
        settings.CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
        settings.CELERY_ALWAYS_EAGER = True


class FastPasswordHasherMixin(object):

    def setup_test_environment(self):
        super().setup_test_environment()
        settings.PASSWORD_HASHERS = ('django.contrib.auth.hashers.MD5PasswordHasher', )


class TemporarySendfileRootMixin(object):

    def setup_test_environment(self):
        super().setup_test_environment()
        settings.SENDFILE_ROOT = tempfile.mkdtemp()



"""
Module related to managing projects with Atlassian's products.
"""

from jira import JIRA

__all__ = ['JiraProject']


class JiraProject(object):
    """A JIRA project class with a simple API leveraging the class :class:`jira.JIRA`."""

    def __init__(self, project=None, server=None, auth=None, feature_type=None):
        self.project = project
        self.server = server
        self.auth = auth
        self.feature_type = feature_type
        self._fields = self._jira = None

    @property
    def fields(self):
        self._fields = self._fields or self.jira.fields()
        return self._fields

    @property
    def features(self):
        count, issues = None, {}
        while count != len(issues):
            count = len(issues)
            issues.update({
                i.id: i for i in self.jira.search_issues(
                    f'project={self.project} AND issuetype="{self.feature_type}"',
                    startAt=count
                )
            })
        return issues.values()

    @property
    def jira(self):
        self._jira = self._jira or JIRA(server=self.server, basic_auth=self.auth)
        return self._jira

    @property
    def versions(self):
        return self.jira.project_versions(self.project)

    def get_field(self, name, fail=True):
        try:
            return next(f for f in self.fields if f['name'] == name)
        except StopIteration:
            if fail:
                raise
        return None

    def get_field_value(self, issue, name, default=None):
        field_id = self.get_field(name)['id']
        field_value = getattr(issue.fields, field_id) or default
        return getattr(field_value, 'value', field_value)



from selenium.webdriver.remote import webelement

from pytoolbox.selenium import common, exceptions

__all__ = ['WebElement']


class WebElement(common.FindMixin, webelement.WebElement):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._specialize()

    @staticmethod
    def clean_value(value):
        return value

    def get_attribute(self, name):
        value = super().get_attribute(name)
        return self.clean_value(value) if name == 'value' else value

    def _specialize(self):
        component = self.get_attribute('data-component')
        if component:
            try:
                self.__class__ = next(  # pylint: disable=invalid-class-object
                    c for c in type(self).__subclasses__() if c.component == component)
            except StopIteration:
                self._specialize_default(component)

    @staticmethod
    def _specialize_default(component):
        raise exceptions.NoSuchSpecializedElementException(
            f'Unable to find a class implementing the component {component}.')



"""
Mix-ins for building your own model managers.
"""

from pytoolbox import module
from pytoolbox.django.models import utils
from pytoolbox.django.models.query import mixins as query_mixins

_all = module.All(globals())

AtomicGetUpdateOrCreateMixin = query_mixins.AtomicGetUpdateOrCreateMixin
AtomicGetRestoreOrCreateMixin = query_mixins.AtomicGetRestoreOrCreateMixin
CreateModelMethodMixin = query_mixins.CreateModelMethodMixin
StateMixin = query_mixins.StateMixin


class RelatedModelMixin(object):

    def get_related_manager(self, field):
        return utils.get_related_manager(self.model, field)

    def get_related_model(self, field):
        return utils.get_related_model(self.model, field)


__all__ = _all.diff(globals())



"""
Some utilities related to the forms.
"""

from copy import copy

from django.contrib import messages
from django.forms.utils import ErrorList

from pytoolbox import module

_all = module.All(globals())


def conditional_required(form, required_dict, data=None, cleanup=False):
    """
    Toggle requirement of some fields based on a dictionary with 'field name' -> 'required boolean'.
    """
    data = data or form.cleaned_data
    for name, value in data.items():
        required = required_dict.get(name, None)
        if required and not value:
            form._errors[name] = ErrorList(['This field is required.'])
        if required is False and cleanup:
            data[name] = None
    return data


def get_instance(form, field_name, request, msg=None):
    """
    Return the instance if the `form` is valid, or try to get it from database.
    Return None if not found and add an error message if set.
    """
    if form.is_valid():
        return form.cleaned_data[field_name]
    queryset = form.fields[field_name].queryset
    try:
        return queryset.get(pk=form.data[form.add_prefix(field_name)])
    except (KeyError, queryset.model.DoesNotExist):
        if msg:
            messages.error(request, msg)


def set_disabled(form, field_name, value=False):
    """Toggle the disabled attribute of a form's field."""
    if value:
        form.fields[field_name].widget.attrs['disabled'] = True
    else:
        try:
            del form.fields[field_name].widget.attrs['disabled']
        except Exception:
            pass


def update_widget_attributes(widget, updates):
    """
    Update attributes of a `widget` with content of `updates` handling classes addition [+],
    removal [-] and toggle [^].

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>> widget = type(str(''), (), {})
    >>> widget.attrs = {'class': 'mondiale'}
    >>> update_widget_attributes(
    ...     widget, {'class': '+pigeon +pigeon +voyage -mondiale -mondiale, ^voyage ^voyageur'})
    >>> asserts.dict_equal(widget.attrs, {'class': 'pigeon voyageur'})
    >>> update_widget_attributes(widget, {'class': '+le', 'cols': 100})
    >>> asserts.dict_equal(widget.attrs, {'class': 'le pigeon voyageur', 'cols': 100})
    """
    updates = copy(updates)
    if 'class' in updates:
        class_set = set([c for c in widget.attrs.get('class', '').split(' ') if c])
        for cls in set([c for c in updates['class'].split(' ') if c]):
            operation, cls = cls[0], cls[1:]
            if operation == '+' or (operation == '^' and cls not in class_set):
                class_set.add(cls)
            elif operation in ('-', '^'):
                class_set.discard(cls)
            else:
                raise ValueError(
                    'updates must be a valid string with "<op>class <op>..." with op in [+-^].')
        widget.attrs['class'] = ' '.join(sorted(class_set))
        del updates['class']
    widget.attrs.update(updates)


def validate_start_end(form, data=None, start_name='start_date', end_name='end_date'):
    """
    Check that the field containing the value of the start field (time, ...) is not bigger (>) than
    the stop.
    """
    data = data or form.cleaned_data
    start, end = data[start_name], data[end_name]
    if start and end and start > end:
        form._errors[end_name] = ErrorList([
            f"The {start_name.replace('_', ' ')} cannot be before the {end_name.replace('_', ' ')}."
        ])


__all__ = _all.diff(globals())



import errno, inspect, http.client, os, re, socket, uuid
from urllib.parse import urlparse

from . import module
from .network.ip import ip_address
from .private import InvalidId, ObjectId

_all = module.All(globals())


class CleanAttributesMixin(object):
    """
    Put validation logic, cleanup code, ... into a method clean_<attribute_name> and this method
    will be called every time the attribute is set.

    **Example usage**

    >>> class Settings(CleanAttributesMixin):
    ...     def __init__(self, locale, broker):
    ...        self.locale = locale
    ...        self.broker = broker
    ...
    ...     def clean_locale(self, value):
    ...         value = str(value)
    ...         assert len(value) == 2
    ...         return value

    >>> settings = Settings('fr', {})
    >>> settings = Settings(10, {})
    >>> assert isinstance(settings.locale, str)
    >>> settings = Settings(100, 'a string')
    Traceback (most recent call last):
        ...
    AssertionError
    """
    def __setattr__(self, name, value):
        cleanup_method = getattr(self, 'clean_' + name, None)
        if cleanup_method:
            value = cleanup_method(value)
        super().__setattr__(name, value)


class StrongTypedMixin(object):
    """
    Annotate arguments of the class __init__ with types and then you'll get a class with type
    checking.

    **Example usage**

    >>> class Settings(StrongTypedMixin):
    ...     def __init__(self, *, locale: (str, list), broker: dict=None, debug: bool=True,
    ...                  timezone=None):
    ...        self.locale = locale
    ...        self.broker = broker
    ...        self.debug = debug
    ...        self.timezone = timezone
    ...
    >>> settings = Settings(locale='fr', broker={}, debug=False)
    >>> settings = Settings(
    ...     locale='fr', broker={}, timezone='this argument is not type checked')
    >>> settings = Settings(locale='fr')
    >>> print(settings.broker)
    None
    >>> settings = Settings(locale=['en', 'fr'], broker={})
    >>> settings = Settings(locale=10, broker={})
    Traceback (most recent call last):
        ...
    AssertionError: Attribute locale must be set to an instance of (<class 'str'>, <class 'list'>)
    """
    def __setattr__(self, name, value):
        the_type = self.__init__.__annotations__.get(name)
        if the_type:
            default = inspect.signature(self.__init__).parameters[name].default
            if value != default:
                assert isinstance(value, the_type), \
                    f'Attribute {name} must be set to an instance of {the_type}'
        super().__setattr__(name, value)


def valid_filename(path):
    """
    Returns True if `path` is a valid file name.

    **Example usage**

    >>> valid_filename('my_file_without_extension')
    False
    >>> valid_filename('my_file_with_extension.mp4')
    True
    """
    try:
        return bool(re.match(r'[^\.]+\.[^\.]+', path))
    except Exception:  # pylint:disable=broad-except
        return False


def valid_ip(address):
    """
    Returns True if `ip` is a valid IP address.

    **Example usage**

    >>> valid_ip('123.0.0.')
    False
    >>> valid_ip('239.232.0.222')
    True
    """
    try:
        ip_address(address)
        return True
    except Exception:  # pylint:disable=broad-except
        return False


def valid_email(email):
    """
    Returns True if `email` is a valid e-mail address.

    **Example usage**

    >>> valid_email('Tabby@croquetes')
    False
    >>> valid_email('Tabby@bernex.ch')
    True
    """
    try:
        return bool(re.match(r'[^@]+@[^@]+\.[^@]+', email))
    except Exception:  # pylint:disable=broad-except
        return False


def valid_int(value):
    """
    Returns True if `value` is a valid integer (can be converted to an int).

    **Example usage**

    >>> valid_int('dimitri is not a valid integer')
    False
    >>> valid_int('-10')
    True
    """
    try:
        int(value)
        return True
    except ValueError:  # pylint:disable=broad-except
        return False


def valid_port(port):
    """
    Returns True if `port` is a valid port.

    **Example usage**

    >>> assert not valid_port(-1)
    >>> assert not valid_port('something not a port')
    >>> assert valid_port('80')
    >>> valid_port(65535)
    True
    """
    try:
        return 0 <= int(port) < 2**16
    except Exception:  # pylint:disable=broad-except
        return False


def valid_secret(secret, none_allowed):
    """
    Returns True if `secret` is a valid secret.

    A valid secret contains at least 8 alpha-numeric characters.

    **Example usage**

    >>> valid_secret('1234', False)
    False
    >>> valid_secret(None, True)
    True
    >>> valid_secret(None, False)
    False
    >>> valid_secret('my_password', False)
    True
    """
    if secret is None and none_allowed:
        return True
    try:
        return bool(re.match(r'[A-Za-z0-9@#$%^&+=-_]{8,}', secret))
    except Exception:  # pylint:disable=broad-except
        return False


def valid_uri(uri, check_404, scheme_mandatory=False, port_mandatory=False, default_port=80,
              excepted_errnos=(errno.ENOENT, errno.ECONNREFUSED, errno.ENETUNREACH), timeout=None):
    """

    *Example usage**

    >>> valid_uri('http://docs.python.org/2/library/httplib.html', check_404=True)
    True
    >>> valid_uri('//docs.python.org/2/library/httplib.html', check_404=True)
    True
    >>> valid_uri('domain_not_exist_404_404/index.htmlw', check_404=True)
    False

    Set default value for port [time'd out request]:

    >>> valid_uri('http://google.com/pytoolbox', check_404=True, default_port=80, timeout=0.2)
    False

    Following the syntax ... in RFC 1808, ... input is presumed ... a path component:

    >>> valid_uri('docs.python.org/2/library/httplib.html', check_404=True)
    False

    This function does not use scheme of `uri` at all, so here is the proof:

    >>> valid_uri('gluster://docs.python.org/2/library/httplib.html', check_404=True)
    True

    Enforce the scheme or the port to being set:

    >>> valid_uri(
    ...     '//domain_not_exist_404_404/index.html:80', check_404=False, scheme_mandatory=True)
    False
    >>> valid_uri('//domain_not_exist_404_404/index.html:80', check_404=False, port_mandatory=True)
    False
    """
    url = urlparse(uri)
    if not url.netloc or scheme_mandatory and not url.scheme or port_mandatory and not url.port:
        return False
    if check_404:
        conn = http.client.HTTPConnection(url.netloc, url.port or default_port, timeout=timeout)
        try:
            conn.request('HEAD', url.path)
            return conn.getresponse().status != 404
        except socket.error as e:
            # Resource does not exist
            if isinstance(e, socket.timeout) or e.errno in excepted_errnos:
                return False
            raise  # Re-raise exception if a different error occurred
        finally:
            conn.close()
    return True


def valid_uuid(value, objectid_allowed=False, none_allowed=False):
    """
    Returns True if `id` is a valid UUID / ObjectId.

    **Example usage**

    >>> valid_uuid(None)
    False
    >>> valid_uuid(None, none_allowed=True)
    True
    >>> valid_uuid('gaga-gogo-gaga-gogo')
    False
    >>> valid_uuid('gaga-gogo-gaga-gogo', objectid_allowed=True)
    False
    >>> valid_uuid(uuid.uuid4(), none_allowed=False)
    True
    >>> valid_uuid(uuid.uuid4().hex, none_allowed=False)
    True
    >>> valid_uuid(str(uuid.uuid4().hex), none_allowed=False)
    True
    >>> valid_uuid(ObjectId())
    False
    >>> valid_uuid(ObjectId(), objectid_allowed=True)
    True
    >>> valid_uuid(ObjectId().binary, objectid_allowed=True)
    True
    """
    if value is None and none_allowed:
        return True
    try:
        uuid.UUID('{{{0}}}'.format(value))  # pylint:disable=consider-using-f-string
    except ValueError:
        if not objectid_allowed:
            return False
        if ObjectId is None:
            raise RuntimeError('bson library not installed')  # pylint:disable=raise-missing-from
        try:
            ObjectId(value)
        except InvalidId:
            return False
    return True


def validate_list(the_list, regexes):
    """
    Validate every element of `the_list` with corresponding regular expression picked-in from
    `regexes`.
    """
    if len(the_list) != len(regexes):
        raise IndexError(
            f'{len(the_list)} elements to validate with '
            f'{len(regexes)} regular expressions')

    for counter, (regex, value) in enumerate(zip(regexes, the_list), 1):
        if not re.match(regex, str(value)):
            raise ValueError(
                f'N°{counter} is invalid:{os.linesep}'
                f'\telement: {value}{os.linesep}'
                f'\tregex:   {regex}')


__all__ = _all.diff(globals())



"""
Pytoolbox's Template tag and filters.
"""

import datetime, os, re

from django import template
from django.conf import settings
from django.template.defaultfilters import stringfilter
from django.templatetags.static import PrefixNode, StaticNode
from django.utils.html import conditional_escape
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

from pytoolbox import humanize, module
from pytoolbox.datetime import secs_to_time as _secs_to_time
from pytoolbox.private import _parse_kwargs_string
from .core import constants

_all = module.All(globals())

try:
    from django.template.defaulttags import include_is_allowed as _include_is_allowed
except ImportError:
    def _include_is_allowed(filepath):
        """
        Removed since Django 1.10
        See commit 04ee4059d71dbc6aa029907e251360eaf00e11bb#diff-45fa5fdd90e8a31a18a1e55ec2f94fa3
        """
        return os.path.abspath(filepath).startswith(settings.STATIC_ROOT)

string_if_invalid = ''  # Official default value
try:  # Django >= 1.8
    string_if_invalid = settings.TEMPLATES['OPTIONS']['string_if_invalid']
except Exception:
    try:  # Django < 1.8
        string_if_invalid = settings.TEMPLATE_STRING_IF_INVALID
    except Exception:
        pass  # Use default


# ====================   =====================   ===============   ===============   =====================
# description            decorator               arguments         input             output
# ====================   =====================   ===============   ===============   =====================
# only accept string     @stringfilter           -                 is a string [1]   return ...
# do not add < > ' " &   is_safe=True            -                 of any type       return ...
# add HTML < > ' " &     needs_autoescape=True   autoescape=None   esc(...) [2]      return mark_safe(...)
# ====================   =====================   ===============   ===============   =====================
#
# [1] Types of string passed to a template code:
#     1. Raw str or unicode. They’re escaped on output if auto-escaping is in effect and presented unchanged, otherwise
#     2. Already marker as safe (SafeBytes/Text, base: SafeData) commonly used for output that contains raw HTML to keep
#     3. Marked as needing escaping, always escaped on output, regardless in autoescape block or not, EscapeBytes/Text
# [2]: esc = conditional_escape if autoescape else lambda x: x

register = template.Library()

NUMERIC_TEST = re.compile(r'^\d+$')
LABEL_TO_CLASS = {
    'ERROR': 'label-important',
    'FAILURE': 'label-important',
    'PENDING': 'label-warning',
    'STARTED': 'label-info',
    'PROGRESS': 'label-info',
    'RETRY': 'label-warning',
    'REVOKED': 'label-inverse',
    'SUCCESS': 'label-success'
}
LABEL_TO_CLASS.update(getattr(settings, 'LABEL_TO_CLASS', {}))


@register.filter(is_safe=True)
def getattribute(value, attribute):
    """
    Gets an attribute of an object dynamically from a string name.

    Source : https://snipt.net/Fotinakis/django-template-tag-for-dynamic-attribute-lookups/

    **Example usage**

    In template::

        {% load pytoolbox_tags %}
        {{ object|getattribute:dynamic_string_var }}
    """
    if hasattr(value, str(attribute)):
        return getattr(value, attribute)
    elif hasattr(value, 'has_key') and attribute in value:
        return value[attribute]
    elif NUMERIC_TEST.match(str(attribute)) and len(value) > int(attribute):
        return value[int(attribute)]
    return string_if_invalid


@register.filter(needs_autoescape=True, safe=True)
@stringfilter
def inline(filepath, msg=True, autoescape=True):
    if filepath in (None, string_if_invalid):
        return string_if_invalid
    if _include_is_allowed(filepath):
        return open(filepath, encoding='utf-8').read()
    if settings.DEBUG and msg:
        filepath_escaped = conditional_escape(filepath) if autoescape else filepath
        return _("[Didn't have permission to include file {0}]").format(filepath_escaped)
    return string_if_invalid


@register.filter(is_safe=True)
def naturalbitrate(bps, kwargs_string=None):
    """
    Return a human readable representation of a bit rate taking `bps` as the rate in bits/s.
    See documentation of :func:`pytoolbox.humanize.naturalbitrate` for further examples.

    Output::

        16487211.33|naturalbitrate -> 16.5 Mb/s
        16487211.33|naturalbitrate:'format={value:.0f} {unit}' -> 16 Mb/s
        16487211.33|naturalbitrate:'format={sign}{value:.0f} {unit}; scale=1' -> 16487 kb/s
        -16487211.33|naturalbitrate:'format={sign}{value:.0f} {unit}' -> -16 Mb/s
        None|naturalbitrate -> (empty string)
        (empty string)|naturalbitrate -> (empty string)
    """
    if bps in (None, string_if_invalid):
        return string_if_invalid
    return humanize.naturalbitrate(
        bps, **_parse_kwargs_string(kwargs_string, format=str, scale=int))


@register.filter(is_safe=True)
def naturalfilesize(the_bytes, kwargs_string=None):
    """
    Return a human readable representation of a *file* size taking `the_bytes` as the size in bytes.
    See documentation of :func:`pytoolbox.humanize.naturalfilesize` for further examples.

    Output::

        16487211.33|naturalfilesize -> 15.7 MB
        16487211.33|naturalfilesize:'system=si' -> 16.5 MiB
        16487211.33|naturalfilesize:'format={value:.0f} {unit}; system=gnu' -> 16 M
        16487211.33|naturalfilesize:'format={sign}{value:.0f} {unit}; scale=1' -> 16101 kB
        -16487211.33|naturalfilesize:'format={sign}{value:.0f} {unit}' -> -16 MB
        None|naturalfilesize -> (empty string)
        (empty string)|naturalfilesize -> (empty string)
    """
    if the_bytes in (None, string_if_invalid):
        return string_if_invalid
    return humanize.naturalfilesize(
        the_bytes, **_parse_kwargs_string(kwargs_string, format=str, scale=int, system=str))


@register.filter(is_safe=True)
def rst_title(value, level):
    """
    Return a title formatted with reSTructuredtext markup.

    * level as number: (1, 2, 3, 4, 5)
    * level as text: ('document', 'subtitle', 'chapter', 'section', 'subsection')

    Output::

        'Document Title'|rst_title:'document' -> ==============\nDocument Title\n==============\n
        'My Subtitle'|rst_title:'subtitle' -> -----------\nMy Subtitle\n-----------\n

    Source: https://snipt.net/Fotinakis/django-template-tag-for-dynamic-attribute-lookups/

    **Example usage**

    In template::

        {% load pytoolbox_tags %}
        {{ 'My chapter'|rst_title:'chapter' }}
    """
    value, level = str(value), str(level)
    length = len(value)
    if level in ('1', 'document'):
        return f"{'=' * length}{os.linesep}{value}{os.linesep}{'=' * length}{os.linesep}"
    elif level in ('2', 'subtitle'):
        return f"{'-' * length}{os.linesep}{value}{os.linesep}{'-' * length}{os.linesep}"
    elif level in ('3', 'chapter'):
        return f"{value}{os.linesep}{'=' * length}{os.linesep}"
    elif level in ('4', 'section'):
        return f"{value}{os.linesep}{'-' * length}{os.linesep}"
    elif level in ('5', 'subsection'):
        return f"{value}{os.linesep}{'~' * length}{os.linesep}"
    return string_if_invalid


@register.filter(is_safe=True)
def secs_to_time(value, defaults_to_zero=False):
    """
    Return an instance of time, taking value as the number of seconds + microseconds
    (e.g. 10.3 = 10s 3000us).

    Output::

        83707.0035|secs_to_time|time:"H:i:s.u" -> 23:15:07.003500
        None|secs_to_time|time:"H:i:s.u"       -> (empty string)
        None|secs_to_time:True|time:"H:i:s.u"  -> 00:00:00.000000
    """
    value = _secs_to_time(value, defaults_to_zero=defaults_to_zero)
    return value if value is not None else string_if_invalid


@register.filter(needs_autoescape=True)
@stringfilter
def status_label(value, autoescape=None, default=''):
    """
    Return the status string represented as a span with a Twitter Bootstrap class.

    Output::

        'ERROR'|status_label -> <span class="label label-important">ERROR</span>

    **Example usage**

    In template::

        {% load pytoolbox_tags %}
        {{ my_status_variable|status_label }}
    """
    esc = conditional_escape if autoescape else lambda x: x
    value = esc(value).upper()
    return mark_safe(f'<span class="label {LABEL_TO_CLASS.get(value, default)}">{value}</span>')


@register.filter(is_safe=True)
def timedelta(value, digits=0):
    """
    Return a string with representation of total seconds given timedelta or a number.

    Output::

        6316.9|timedelta -> 01:45:17
        datetime.timedelta(0, 6317)|timedelta -> 01:45:17
        datetime.timedelta(days=10, hours=2)|timedelta -> 10 days, 02:00:00
        datetime.timedelta(days=10, seconds=30.408)|timedelta -> 10 days, 00:00:30
        datetime.timedelta(days=10, seconds=30.408)|timedelta:2 -> 10 days, 00:00:30.45
        None|timedelta:10 -> (empty string)
        (empty string)|timedelta -> (empty string)
    """
    if value in (None, string_if_invalid):
        return string_if_invalid
    seconds = value.total_seconds() if hasattr(value, 'total_seconds') else float(value)
    return force_text(datetime.timedelta(seconds=round(seconds, digits))).replace('days', _('days'))


@register.filter
def verbose_name(instance):
    """Return the verbose name (singular) of a model."""
    if instance in (None, string_if_invalid):
        return string_if_invalid
    return constants.DEFFERED_REGEX.sub('', force_text(instance._meta.verbose_name))


@register.filter
def verbose_name_plural(instance):
    """Return the verbose name (plural) of a model."""
    if instance in (None, string_if_invalid):
        return string_if_invalid
    return constants.DEFFERED_REGEX.sub('', force_text(instance._meta.verbose_name))


# TAGS ---------------------------------------------------------------------------------------------

class StaticPathNode(StaticNode):

    @classmethod
    def handle_simple(cls, path):
        return os.path.join(PrefixNode.handle_simple('STATIC_ROOT'), path)


@register.tag('static_abspath')
def static_abspath(parser, token):
    """
    Joins the given path with the STATIC_ROOT setting.

    Usage::

        {% static path [as varname] %}

    Examples::

        {% static "myapp/css/base.css" %}
        {% static variable_with_path %}
        {% static "myapp/css/base.css" as admin_base_css %}
        {% static variable_with_path as varname %}

    """
    return StaticPathNode.handle_token(parser, token)


__all__ = _all.diff(globals())
